"""
This contains all the statistical functions.
"""

# pylint: disable=too-many-arguments

import logging
from typing import Callable, List, Optional, Tuple, Union

import numpy as np

from yupi._checkers import (
    check_exact_dim,
    check_same_dim,
    check_same_dt,
    check_same_t,
    check_uniform_time_spaced,
)
from yupi.trajectory import Trajectory
from yupi.transformations import subsample
from yupi.vector import Vector


def collect_at_step(
    trajs: List[Trajectory],
    step: int,
    warnings: bool = True,
    velocity: bool = False,
    func: Optional[Callable[[Vector], Vector]] = None,
) -> np.ndarray:
    """
    Collects the positional data (or velocity) of each trajectory at a given
    step.

    Parameters
    ----------
    trajs : List[Trajectory]
        List of trajectories.
    step : int
        Index of the collected vector of each trajectory.
    warnings : bool
        If True, warns if the trajectory is shorter than the step, by default
        True.
    velocity : bool
        If True, the velocity of the trajectory is used, by default False.
    func : Optional[Callable[[Vector], Vector]]
        Function to apply to the collected vector of each trajectory.
        By default, the identity function.

    Returns
    -------
    np.ndarray
        Array of collected data.

    See Also
    --------
    collect_at_time, collect_step_lagged, collect_time_lagged, collect
    """
    return collect(trajs, at=int(step), warnings=warnings, velocity=velocity, func=func)


def collect_at_time(
    trajs: List[Trajectory],
    time: float,
    warnings: bool = True,
    velocity: bool = False,
    func: Optional[Callable[[Vector], Vector]] = None,
) -> np.ndarray:
    """
    Collects the positional data (or velocity) of each trajectory at a given
    time.

    Parameters
    ----------
    trajs : List[Trajectory]
        List of trajectories.
    time : float
        Time of the collected vector of each trajectory.

        It is calculated using the trajectory's dt.
    warnings : bool
        If True, warns if the trajectory is shorter than the time, by default
        True.
    velocity : bool
        If True, the velocity of the trajectory is used, by default False.
    func : Optional[Callable[[Vector], Vector]]
        Function to apply to the collected vector of each trajectory.
        By default, the identity function.

    Returns
    -------
    np.ndarray
        Array of collected data.

    See Also
    --------
    collect_at_step, collect_step_lagged, collect_time_lagged, collect
    """
    return collect(
        trajs, at=float(time), warnings=warnings, velocity=velocity, func=func
    )


def collect_step_lagged(
    trajs: List[Trajectory],
    step: int,
    warnings: bool = True,
    velocity: bool = False,
    concat: bool = True,
    func: Optional[Callable[[Vector], Vector]] = None,
) -> np.ndarray:
    """
    Collects the positional data (or velocity) of each trajectory lagged by a
    given step.

    Parameters
    ----------
    trajs : List[Trajectory]
        List of trajectories.
    step : int
        Number of steps to lag.
    warnings : bool
        If True, warns if the trajectory is shorter than the step, by default
        True.
    velocity : bool
        If True, the velocity of the trajectory is used, by default False.
    concat : bool
        If True, the data is concatenated, by default True.
    func : Optional[Callable[[Vector], Vector]]
        Function to apply to the collected vector of each trajectory.
        By default, the identity function.

    Returns
    -------
    np.ndarray
        Array of collected data.

    See Also
    --------
    collect_at_step, collect_at_step, collect_time_lagged, collect
    """
    return collect(
        trajs,
        lag=int(step),
        concat=concat,
        warnings=warnings,
        velocity=velocity,
        func=func,
    )


def collect_time_lagged(
    trajs: List[Trajectory],
    time: float,
    warnings: bool = True,
    velocity: bool = False,
    concat: bool = True,
    func: Optional[Callable[[Vector], Vector]] = None,
) -> np.ndarray:
    """
    Collects the positional data (or velocity) of each trajectory lagged by a
    given time.

    Parameters
    ----------
    trajs : List[Trajectory]
        List of trajectories.
    time : float
        Time to lag.
    warnings : bool
        If True, warns if the trajectory is shorter than the step, by default
        True.
    velocity : bool
        If True, the velocity of the trajectory is used, by default False.
    concat : bool
        If True, the data is concatenated, by default True.
    func : Optional[Callable[[Vector], Vector]]
        Function to apply to the collected vector of each trajectory.
        By default, the identity function.

    Returns
    -------
    np.ndarray
        Array of collected data.

    See Also
    --------
    collect_at_step, collect_at_time, collect_step_lagged, collect
    """
    return collect(
        trajs,
        lag=float(time),
        concat=concat,
        warnings=warnings,
        velocity=velocity,
        func=func,
    )


def collect(
    trajs: List[Trajectory],
    lag: Optional[Union[int, float]] = None,
    concat: bool = True,
    warnings: bool = True,
    velocity: bool = False,
    func: Optional[Callable[[Vector], Vector]] = None,
    at: Optional[Union[int, float]] = None,  # pylint: disable=invalid-name
) -> np.ndarray:
    """
    Collect general function.

    It can collect the data of each trajectory lagged by a given step or time
    (step if ``lag`` is ``int``, time if ``lag`` is ``float``). It can also
    collect the data of each trajectory at a given step or time (step if ``at``
    is ``int``, time if ``at`` is ``float``). Both ``lag`` and ``at``
    parameters can not be used at the same time.

    Parameters
    ----------
    trajs : List[Trajectory]
        Group of trajectories.
    lag : Optional[Union[int, float]]
        If int, the number of samples to lag. If float, the time to lag.
    concat : bool, optional
        If true each trajectory stracted data will be concatenated in
        a single array, by default True.
    warnings : bool, optional
        If true, warnings will be printed if a trajectory is shorter
        than the lag, by default True.
    velocity : bool, optional
        If true, the velocity will be returned (calculated using the
        lag if given), by default False.
    func : Optional[Callable[[Vector], Vector]]
        Function to apply to each resulting vector, by default None.
    at : Optional[Union[int, float]]
        If int, the index of the collected vector in the trajectory. If
        float, it is taken as time and the index is calculated using
        the trajectory's dt.

    Returns
    -------
    np.ndarray
        Collected data.

    Raises
    ------
    ValueError
        If ``lag`` and ``at`` are given at the same time.
    """

    checks = [
        isinstance(lag, int),
        isinstance(lag, float),
        isinstance(at, int),
        isinstance(at, float),
    ]

    if sum(checks) == 0:
        lag = 0
        checks[0] = True
    if sum(checks) > 1:
        raise ValueError(
            "You can not set `lag` and `at` parameters at the " "same time"
        )
    is_lag = checks[0] or checks[1]

    data = []
    for traj in trajs:
        if is_lag:
            assert lag is not None
            step = int(lag / traj.dt) if checks[1] else int(lag)
        else:
            assert at is not None
            step = int(at / traj.dt) if checks[3] else int(at)

        current_vec = traj.r
        if step == 0:
            if velocity:
                current_vec = traj.v
            if func is not None:
                current_vec = func(current_vec)
            data.append(current_vec if is_lag else current_vec[step])
            continue

        if warnings and step >= len(current_vec):
            logging.warning(
                "Trajectory %s is shorten than %i samples", traj.traj_id, step
            )
            continue

        if not is_lag:  # Is at
            data.append(current_vec[step])
            continue

        lagged_vec = current_vec[step:] - current_vec[:-step]
        if velocity:
            lagged_vec /= traj.dt * step

        if func is not None:
            lagged_vec = func(lagged_vec)

        data.append(lagged_vec)

    if concat and is_lag:
        return np.concatenate(data)
    equal_len = np.all([len(d) == len(data[0]) for d in data])
    return np.array(data) if equal_len else np.array(data, dtype=object)


@check_same_dt
@check_exact_dim(2)
@check_uniform_time_spaced
def turning_angles_ensemble(
    trajs: List[Trajectory], accumulate=False, degrees=False, centered=False, wrap=True
) -> np.ndarray:
    """
    Return a concatenation of all the turning angles that forms
    a set of trajectories.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    accumulate : bool, optional
        If True, turning angles are measured with respect to an axis
        define by the initial velocity (i.e., angles between initial
        and current velocity). Otherwise, relative turning angles
        are computed (i.e., angles between succesive velocity vectors).
        By default False.
    degrees : bool, optional
        If True, angles are given in degrees. Otherwise, the units
        are radians. By default False.
    centered : bool, optional
        If True, angles are wrapped on the interval ``[-pi, pi]``.
        Otherwise, the interval ``[0, 2*pi]`` is chosen. By default
        False.

    Returns
    -------
    np.ndarray
        Concatenated array of turning angles for a list of Trajectory
        objects.
    """

    theta = [t.turning_angles(accumulate, degrees, centered, wrap) for t in trajs]
    return np.concatenate(theta)


@check_same_dim
def speed_ensemble(trajs: List[Trajectory], step: int = 1) -> np.ndarray:
    """
    Estimate speeds of the list of trajectories, ``trajs``,
    by computing displacements according to a certain sample
    frequency given by ``step``.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    step : int
        Numer of sample points.

    Returns
    -------
    np.array
        Concatenated array of speeds.
    """

    trajs_ = [subsample(traj, step) for traj in trajs]
    return np.concatenate([traj.v.norm for traj in trajs_])


@check_same_t
def msd_ensemble(trajs: List[Trajectory]) -> np.ndarray:
    """
    Compute the square displacements for every Trajectory object
    stored in ``trajs`` as the square of the current position vector
    that has been subtracted the initial position.

    Trajectories should have the same length.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.

    Returns
    -------
    np.ndarray
        Array of square displacements with shape ``(n, N)``, where
        ``n`` is the total number of time steps and ``N`` the number
        of trajectories.
    """

    _msd = []
    for traj in trajs:
        # Position vectors
        r = traj.r

        # Square displacements
        r_2 = (r - r[0]) ** 2  # Square coordinates
        r_2_dis = np.sum(r_2, axis=1)  # Square distances
        _msd.append(r_2_dis)  # Append square distances

    # Transpose to have time/trials as first/second axis
    return np.transpose(_msd)


@check_same_dt
@check_uniform_time_spaced
def msd_time(trajs: List[Trajectory], lag: int) -> np.ndarray:
    """
    Estimate the mean square displacement for every Trajectory
    object stored in ``trajs`` as the average of the square of
    dispacement vectors as a function of the lag time.

    This is a convenience estimator specially when trajectories
    do not have equal lengths.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    lag : int
        Number of steps that multiplied by ``dt`` defines the lag
        time.

    Returns
    -------
    np.ndarray
        Array of mean square displacements with shape ``(lag, N)``,
        where ``N`` the number of trajectories.
    """

    _msd = []
    for traj in trajs:
        # Position vectors
        r = traj.r

        # Compute msd for a single trajectory
        current_msd = np.empty(lag)
        for lag_ in range(1, lag + 1):
            # Lag displacement vectors
            lagged_r = r[lag_:] - r[:-lag_]
            # Lag displacement
            dr2 = np.sum(lagged_r**2, axis=1)
            # Averaging over a single realization
            current_msd[lag_ - 1] = np.mean(dr2)

        # Append all square displacements
        _msd.append(current_msd)

    # Transpose to have time/trials as first/second axis
    return np.transpose(_msd)


@check_same_dim
def msd(
    trajs: List[Trajectory], time_avg: bool = True, lag: Optional[int] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Estimate the mean square displacement of the list of Trajectory
    objects, ``trajs``, providing the options of averaging over the
    ensemble of realizations or over time.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    time_avg : bool, optional
        If True, mean square displacement is estimated averaging over
        time. Otherwise, an ensemble average will be performed and all
        Trajectory objects will have to have the same length. By default
        True.
    lag : Optional[int]
        If None, ``time_avg`` should be set to ``False`` indicating
        ensemble average. Otherwise, ``lag`` is taken as the number
        of steps that multiplied by ``dt`` defines the lag time. By
        default None.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Tuple containing the array of mean square displacements and
        the array of standard deviations.
    """

    if not time_avg:
        _msd = msd_ensemble(trajs)  # Ensemble average
    elif lag is None:
        raise ValueError("You must set 'lag' param if 'time_avg' is True")
    else:
        _msd = msd_time(trajs, lag)  # Time average

    msd_mean = np.mean(_msd, axis=1)  # Mean
    msd_std = np.std(_msd, axis=1)  # Standard deviation
    return msd_mean, msd_std


@check_same_t
def vacf_ensemble(trajs: List[Trajectory]) -> np.ndarray:
    """
    Compute the pair-wise dot product between initial and current
    velocity vectors for every Trajectory object stored in ``trajs``.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.

    Returns
    -------
    np.ndarray
        Array of velocity dot products with shape ``(n, N)``, where
        ``n`` is the total number of time steps and ``N`` the number
        of trajectories.
    """

    _vacf = []
    for traj in trajs:
        # Cartesian velocity components
        v = traj.v

        # Pair-wise dot product between velocities at t0 and t
        v0_dot_v = np.sum(v[0] * v, axis=1)

        # Append all veloctiy dot products
        _vacf.append(v0_dot_v)

    # Transpose to have time/trials as first/second axis
    return np.transpose(_vacf)


@check_same_dt
@check_uniform_time_spaced
def vacf_time(trajs: List[Trajectory], lag: int) -> np.ndarray:
    """
    Estimate the velocity autocorrelation function for every
    Trajectory object stored in ``trajs`` as the average of the
    dot product between velocity vectors that are distant a certain
    lag time.

    This is a convenience estimator specially when trajectories do
    not have equal lengths.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    lag : int
        Number of steps that multiplied by ``dt`` defines the lag
        time.

    Returns
    -------
    np.ndarray
        Array of velocity autocorrelation function with shape
        ``(lag, N)``, where ``N`` is the number of trajectories.
    """

    _vacf = []
    for traj in trajs:
        # Cartesian velocity components
        v = traj.v

        # Compute vacf for a single trajectory
        current_vacf = np.empty(lag)
        for lag_ in range(1, lag + 1):
            # Multiply components given lag
            v1, v2 = v[:-lag_], v[lag_:]
            v1v2 = (v1 - v1.mean(axis=0)) * (v2 - v2.mean(axis=0))

            # Dot product for a given lag time
            v1_dot_v2 = np.sum(v1v2, axis=1)

            # Averaging over a single realization
            current_vacf[lag_ - 1] = np.mean(v1_dot_v2)

        # Append the vacf for a every single realization
        _vacf.append(current_vacf)

    # Aranspose to have time/trials as first/second axis
    return np.transpose(_vacf)


@check_same_dim
def vacf(
    trajs: List[Trajectory], time_avg: bool = True, lag: Optional[int] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Estimate the velocity autocorrelation function of the list of
    Trajectory objects, ``trajs``, providing the options of averaging
    over the ensemble of realizations or over time.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    time_avg : bool, optional
        If True, velocity autocorrelation function is estimated
        averaging over time. Otherwise, an ensemble average will be
        performed and all Trajectory objects will have to have the
        same length. By default True.
    lag : Optional[int]
        If None, ``time_avg`` should be set to ``False`` indicating
        ensemble average. Otherwise, ``lag`` is taken as the number
        of steps that multiplied by ``dt`` defines the lag time.
        By default None.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Tuple containing the array of velocity autocorrelation function
        and the array of standard deviations.
    """

    if not time_avg:
        _vacf = vacf_ensemble(trajs)  # Ensemble average
    elif lag is None:
        raise ValueError("You must set 'lag' param if 'time_avg' is True")
    else:
        _vacf = vacf_time(trajs, lag)  # Time average

    vacf_mean = np.mean(_vacf, axis=1)  # Mean
    vacf_std = np.std(_vacf, axis=1)  # Standard deviation
    return vacf_mean, vacf_std


def _kurtosis(arr):
    """
    Compute the kurtosis of the array, `arr`.

    If `arr` is not a one-dimensional array, it should
    be a horizontal collection of column vectors.

    Parameters
    ----------
    arr : np.adarray
        Data for which the kurtosis is calculated.

    Returns
    -------
    float
        Kurtosis of the data set.
    """

    arr = np.squeeze(arr)

    # ONE-DIMENSIONAL CASE
    if len(arr.shape) == 1:
        # Subtract the mean position at every time instant
        arr_zm = arr - arr.mean()

        # Second and fourth central moments averaging
        # over repetitions
        m_2 = np.mean(arr_zm**2)
        m_4 = np.mean(arr_zm**4)

        # Compute kurtosis for those cases in which the
        # second moment is different from zero
        if m_2 == 0:
            return 0
        kurt = m_4 / m_2**2
        return kurt

    # MULTIDIMENSIONAL CASE
    # arr should have shape (dim, trials)
    # (i.e., a horizontal sequence of column vectors)

    # Subtract the mean position
    arr_zm = arr - arr.mean(1)[:, None]

    try:
        # Inverse of the estimated covariance matrix
        cov_inv = np.linalg.inv(np.cov(arr))
    except np.linalg.LinAlgError:
        # Exception for the case of singular matrices
        return 0

    # Kurtosis definition for multivariate r.v.'s
    _k = np.sum(arr_zm * (cov_inv @ arr_zm), axis=0)
    kurt = np.mean(_k**2)

    return kurt


@check_same_t
def kurtosis_ensemble(trajs: List[Trajectory]) -> np.ndarray:
    """Estimate kurtosis as a function of time of the
    list of Trajectory objects, ``trajs``. The average
    is perform over the ensemble of realizations.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.

    Returns
    -------
    np.ndarray
        Kurtosis at every time instant.
    """

    # Get ensemble positions where axis 0/1/2 are
    # in the order trials/time/dim
    r = Vector([traj.r for traj in trajs])

    # Set trials as the last axis
    moved_r = np.moveaxis(r, 0, 2)

    # Compute kurtosis at every time instant (loop over time)
    kurt = [_kurtosis(r_) for r_ in moved_r]

    return np.array(kurt)


@check_same_dt
@check_uniform_time_spaced
def kurtosis_time(trajs: List[Trajectory], lag: int) -> np.ndarray:
    """
    Estimate the kurtosis for every Trajectory object stored
    in ``trajs``.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    lag : int
        Number of steps that multiplied by ``dt`` defines the lag
        time.

    Returns
    -------
    np.ndarray
        Array of velocity autocorrelation function with shape
        ``(lag, N)``, where ``N`` is the number of trajectories.
    """

    kurt = []
    for traj in trajs:
        current_kurt = np.empty(lag)
        for lag_ in range(lag):
            try:
                lagged_r = traj.r[lag_:] - traj.r[:-lag_]
            except ValueError:
                current_kurt[lag_] = 0
                continue
            current_kurt[lag_] = _kurtosis(lagged_r.T)
        kurt.append(current_kurt)
    return np.transpose(kurt)


@check_same_dim
def kurtosis(
    trajs: List[Trajectory], time_avg: bool = True, lag: Optional[int] = None
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Estimate the kurtosis of the list of Trajectory objects, ``trajs``,
    providing the options of averaging over the ensemble of realizations
    or over time.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    time_avg : bool, optional
        If True, kurtosis is estimated averaging over time. Otherwise,
        an ensemble average will be performed and all Trajectory objects
        will have to have the same length. By default True.
    lag : Optional[int]
        If None, ``time_avg`` should be set to ``False`` indicating
        ensemble average. Otherwise, ``lag`` is taken as the number
        of steps that multiplied by ``dt`` defines the lag time.
        By default None.

    Returns
    -------
    Tuple[np.ndarray, Optional[np.ndarray]]
        Tuple containgin the kurtosis and the standar deviations.
    """

    if not time_avg:
        return kurtosis_ensemble(trajs), None
    if lag is None:
        raise ValueError("You must set 'lag' param if 'time_avg' is True")

    kurt = kurtosis_time(trajs, lag)
    kurt_mean = np.mean(kurt, axis=1)
    kurt_std = np.std(kurt, axis=1)
    return kurt_mean, kurt_std


@check_same_dim
def kurtosis_reference(trajs: List[Trajectory]) -> float:
    """Get the sampled kurtosis for the case of
    ``len(trajs)`` trajectories whose position
    vectors are normally distributed.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input trajectories.

    Returns
    -------
    float
        Value of kurtosis.
    """

    dim = trajs[0].dim
    count = len(trajs)
    kurt = dim * (dim + 2)
    if count == 1:
        return kurt
    return kurt * (count - 1) / (count + 1)


@check_same_dt
@check_uniform_time_spaced
def psd(
    trajs: List[Trajectory], lag: int, omega: bool = True
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Estimate the power spectral density of a list of Trajectory object
    as the Fourier transform of its velocity autocorrelation function.

    Parameters
    ----------
    trajs : List[Trajectory]
        Input list of trajectories.
    lag : int
        Number of steps that multiplied by ``dt`` defines the lag
        time.
    omega: bool
        If True, return the angular frequency instead of the frequency.

    Returns
    -------
    np.ndarray, np.ndarray, np.ndarray
        Power spectral density mean, standard deviation, and
        frequency axis.
    """

    _vacf = vacf_time(trajs, lag)
    _ft = np.fft.fft(_vacf, axis=0) * trajs[0].dt
    _ft = np.fft.fftshift(_ft)
    ft_abs = np.abs(_ft)
    ft_mean = np.mean(ft_abs, axis=1)
    ft_std = np.std(ft_abs, axis=1)

    frec = 2 * np.pi * np.fft.fftfreq(lag, trajs[0].dt)
    frec = np.fft.fftshift(frec)

    return ft_mean, ft_std, frec * 2 * np.pi if omega else frec



"""
This module contains a set of functions capable of applying
transformations to trajectories such as filtering, resampling, etc.

All the resources of this module should be imported directly
from ``yupi.transormations``.
"""

from yupi.transformations._filters import exp_convolutional_filter
from yupi.transformations._resamplers import (
    resample,
    subsample,
)
from yupi.transformations._transformations import add_moving_FoR

__all__ = [
    "subsample",
    "exp_convolutional_filter",
    "add_moving_FoR",
    "resample",
]



import logging
from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple

import cv2
import numpy as np

from yupi.tracking.algorithms import TrackingAlgorithm, _resize_frame
from yupi.tracking.undistorters import Undistorter
from yupi.trajectory import Trajectory
from yupi.transformations import add_moving_FoR
from yupi.transformations._affine_estimator import AffineParams, _get_affine

# pylint: disable=protected-access


Centroid = Tuple[int, int]
"""Centroid of a tracked object: x, y."""

Bounds = Tuple[int, int, int, int]
"""Bounds of a frame: x_min, x_max, y_min, y_max."""


class ROI:
    """
    Region of interest.

    Region that can be tracked by the algorithms throughout the sequence
    of image frames.

    Parameters
    ----------
    size : Tuple[float, float]
        Size of the region of interest.

        If both tuple's values are grater than 1 then they are rounded
        and taken as pixels. Otherwise, if both values are less than 1,
        the size is taken relative to the video frame size.
    init_mode : str, optional
        ROI's initialization mode, by default 'manual'.

        Defines the way ROI initial position is setted.

        The ``init_mode`` parameter can be manual or center. These
        modes are stored in ``ROI.MANUAL_INIT_MODE`` and
        ``ROI.CENTER_INIT_MODE``.
    scale : float, optional
        Scale of the sample frame to set ROI initial position if
        ``init_method`` is set to ``'manual'``, by default 1.

    Attributes
    ----------
    width : float
        Width of the ROI.

        If the width value is between 0 and 1 then this is taken
        relative to the frames. Otherwise it is a rounded value and
        taken as pixels.
    height : float
        Height of the ROI.

        If the height value is between 0 and 1 then this is taken
        relative to the frames. Otherwise it is a rounded value and
        taken as pixels.
    init_mode : str
        ROI's initialization mode.
    scale : float
        Scale of the sample frame to set ROI initial position if
        ``init_method`` is set to ``'manual'``.

    Examples
    --------
    >>> ROI((120, 120), ROI.MANUAL_INIT_MODE)
    ROI: size=(120, 120) init_mode=manual scale=1

    Raises
    ------
    ValueError
        If any size value is negative.
    ValueError
        If one of the size value is grater than 1 and the other is less
        than 1.
    ValueError
        If ROI initialization mode is neither ``'manual'`` or
        ``'center'``.
    """

    MANUAL_INIT_MODE = "manual"
    """Manual initialization mode for the ROI"""
    CENTER_INIT_MODE = "center"
    """Center initialization mode for the ROI"""

    def __init__(
        self,
        size: Tuple[float, float],
        init_mode: str = MANUAL_INIT_MODE,
        scale: float = 1,
    ):

        if size[0] <= 0 or size[1] <= 0:
            raise ValueError("ROI's size values must be positives")

        # TODO: check for a more pythonic way to do this comprobation
        if (size[0] < 1 and size[1] > 1) or (size[0] > 1 and size[1] < 1):
            raise ValueError(
                "Size values must be between 0 and 1 both or "
                "integers greater than 0 both"
            )

        if init_mode not in (ROI.CENTER_INIT_MODE, ROI.MANUAL_INIT_MODE):
            raise ValueError(f"ROI '{init_mode}' initialization mode unknown")

        if scale < 0:
            raise ValueError("ROI scale must be non negative")

        self.width, self.height = size
        self.init_mode = init_mode
        self.scale = scale

        self._prev_centroid: Centroid
        self._centroid: Centroid
        self._global_height: int
        self._global_width: int

    def __repr__(self):
        return (
            "ROI: size=({self.width}, {self.height}) "
            "init_mode={self.init_mode} scale={self.scale}"
        )

    def _recenter(self, centroid: Optional[Centroid]) -> None:
        """
        Recenters ROI position.

        Parameters
        ----------
        centroid : tuple of int
            New center of the ROI.
        """

        # Get the centroid refered to the roi
        if centroid is not None:
            roi_cx, roi_cy = centroid

            # Get the centroid refered to the full image
            c_x = self._prev_centroid[0] - int(self.width / 2) + roi_cx
            c_y = self._prev_centroid[1] - int(self.height / 2) + roi_cy

            c_x = min(c_x, self._global_width)
            c_x = max(c_x, 0)
            c_y = min(c_y, self._global_height)
            c_y = max(c_y, 0)

            self._centroid = (c_x, c_y)

        else:
            self._centroid = self._prev_centroid

    def _get_bounds(self, prev: bool = False) -> Bounds:
        """
        ROI's bounds.

        Calculates the ROI's bounds according to its center, width,
        height and the global bounds.

        Parameters
        ----------
        prev : bool
            Use previous roi center instead of current

        Returns
        -------
        xmin : int
            Mnimum bound on X axis.
        xmax : int
            Maximun bound on X axis.
        ymin : int
            Mnimum bound on Y axis.
        ymax : int
            Maximum bound on Y axis.
        """

        if prev:
            c_x, c_y = self._prev_centroid
        else:
            c_x, c_y = self._centroid

        half_width, half_height = int(self.width / 2), int(self.height / 2)
        xmin = max(c_x - half_width, 0)
        xmax = min(c_x + half_width, self._global_width)
        ymin = max(c_y - half_height, 0)
        ymax = min(c_y + half_height, self._global_height)
        return xmin, xmax, ymin, ymax

    def _center_init(self, frame: np.ndarray) -> Centroid:
        """
        Initialize ROI using center initialization mode.

        Parameters:
        frame : np.ndarray
            Frame used as reference to initialize ROI position at its
            center.

        Returns
        -------
        tuple of int
            Center of the ROI.
        """

        self._global_height, self._global_width = frame.shape[:2]
        self._centroid = self._global_width // 2, self._global_height // 2
        return self._centroid

    # TODO: check for 'win2_name' utility. Maybe it should be 'ROI' as
    # Default so there is no need to pass it as a parameter
    def _manual_init(self, frame: np.ndarray, name: str) -> Centroid:
        """
        Initialize ROI using manual initialization mode.

        Parameter
        ---------
        frame : np.ndarray
            Frame used as reference to initialize ROI position manually.
        name : str
            Name of the tracking object.

        Returns
        -------
        tuple of int
            Center of the ROI.
        """

        win1_name = (
            "Initialization of trackers: Click on the initial "
            f"position of: {name.upper()}"
        )
        logging.info("Open the video window to select %s's center", name)

        self._global_height, self._global_width = frame.shape[:2]

        frame_ = _resize_frame(frame, scale=self.scale)
        cv2.imshow(win1_name, frame_)

        roi_initialized = False
        # Callback handler to manually set the roi
        def on_click(event, x, y, flags, param):
            if event == cv2.EVENT_LBUTTONDOWN:
                # Global roi center coordinates
                self._centroid = int(x / self.scale), int(y / self.scale)

                # Copy of true frame and its resized version
                img_ = frame_.copy()

                # Draw a circle in the selected pixel
                cv2.circle(img_, (x, y), 3, (0, 255, 255), 1)
                xmin, xmax, ymin, ymax = self._get_bounds()
                pt1 = (int(xmin * self.scale), int(ymin * self.scale))
                pt2 = (int(xmax * self.scale), int(ymax * self.scale))
                cv2.rectangle(img_, pt1, pt2, (0, 255, 255), 1)

                msg = "ROI initialized. Press any key to continue."

                cv2.putText(
                    img_, msg, (10, 20), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 255, 255), 1
                )

                cv2.imshow(win1_name, img_)

                nonlocal roi_initialized
                roi_initialized = True
                logging.info("ROI initialized. Press any key to continue.")

        cv2.setMouseCallback(win1_name, on_click)
        while not roi_initialized:
            cv2.waitKey(0)
            logging.info(
                "Waiting for ROI initialization. Please "
                "click on the center of the %s's ROI on "
                "the video window.",
                name,
            )

        return self._centroid

    def _check_roi_init(self, name: str) -> bool:
        """
        Checks for ROI initialization.

        Parameter
        ---------
        name : str
            Name of the tracking object.

        Returns
        -------
        bool
            Whether or not the ROI is initialized.
        """

        if not self._prev_centroid[0]:
            logging.error("ROI was not initialized in %s", name)
            return False

        cv2.destroyAllWindows()
        logging.info("ROI initialized in %s", name)
        return True

    def _initialize(self, name: str, first_frame: np.ndarray) -> bool:
        """
        Initialize ROI.

        Parameters
        ----------
        name : str
            Name of the tracking object.
        first_frame : np.ndarray
            First frame of the video.

            If ROI's initialization mode is set to ``'manual'`` this
            frame will be shown to select the tracking object center.

        Returns
        -------
        bool
            Whether or not the ROI was initialized.
        """

        height, weight = first_frame.shape[:2]
        if self.width <= 1:
            self.width *= weight
        if self.height <= 1:
            self.height *= height

        # Initialize ROI coordinates manually by user input
        if self.init_mode == ROI.MANUAL_INIT_MODE:
            self._centroid = self._manual_init(first_frame, name)
        else:
            self._centroid = self._center_init(first_frame)

        self._prev_centroid = self._centroid
        return self._check_roi_init(name)

    def _crop(self, frame: np.ndarray, prev: bool = False) -> np.ndarray:
        """
        Crops a frame according to the ROI's bounds.

        Parameters
        ----------
        frame : np.ndarray
            Frame that will be cropped.
        prev : bool
            Use previous roi center instead of current

        Returns
        -------
        window : np.ndarray
            Cropped part of the frame.
        """

        self._global_height, self._global_width = frame.shape[:2]
        # Bounds of the roi
        xmin, xmax, ymin, ymax = self._get_bounds(prev)
        window = frame[ymin:ymax, xmin:xmax, :]
        return window.copy()


class ObjectTracker:
    """
    Tracks an object inside a ROI according to a tracking algorithm.

    Parameters
    ----------
    name : str
        Name of the tracked object.
    algorithm : TrackingAlgorithm
        Algorithm used to track the object.
    roi : ROI
        Region of interest where the object will be tracked.
    preprocessing : Callable[[np.ndarray], np.ndarray], optional
        Preprocessing function aplied to the frame before being used by
        the algorithm.

    Attributes
    ----------
    name : str
        Name of the tracked object.
    algorithm : TrackingAlgorithm
        Algorithm used to track the object.
    roi : ROI
        Region of interest where the object will be tracked.
    history : List[Centroid]
        ROI's position in every frame of the video.
    preprocessing : Optional[Callable[[np.ndarray], np.ndarray]]
        Preprocessing function aplied to the frame before being used by
        the algorithm.
    """

    def __init__(
        self,
        name: str,
        algorithm: TrackingAlgorithm,
        roi: ROI,
        preprocessing: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ):
        self.name = name
        self.roi = roi
        self.algorithm = algorithm
        self.preprocessing = preprocessing
        self.history: List[Centroid] = []
        self.mask: np.ndarray

    def _init_roi(self, frame: np.ndarray) -> bool:
        return self.roi._initialize(self.name, frame)

    def _track(self, frame: np.ndarray):
        """
        Tracks the center of the object.

        Given a new frame, the center of the object inside the ROI is
        recalculated using the selected algorithm.

        Parameters
        ----------
        frame : np.ndarray
            Frame used by the algorithm to detect the tracked object's
            new center.
        """

        # Get only the ROI from the current frame
        roi_bound = self.roi._get_bounds()

        # Detect the object using the tracking algorithm
        self.mask, centroid = self.algorithm.detect(
            frame, roi_bound, self.preprocessing
        )

        # Update the roi center using current ant coordinates
        self.roi._recenter(centroid)

        # Update data
        self.history.append(self.roi._centroid)


class CameraTracker:
    """
    Tracks the camera movement.

    Parameters
    ----------
    roi : ROI
        Region of interest where the background changes will be
        detected.

    Attributes
    ----------
    roi : ROI
        Region of interest where the background changes will be
        detected.
    affine_params_history : List[AffineParams]
        History of all the affine parameters
    """

    def __init__(self, roi: ROI):
        self.affine_params_history: List[AffineParams] = []
        self.mse: List[float] = []
        self.roi = roi
        self.features: Any

    def _init_roi(self, prev_frame: np.ndarray) -> bool:
        return self.roi._initialize("Camera", prev_frame)

    # Track the floor
    def _track(
        self, prev_frame: np.ndarray, frame: np.ndarray, ignored_regions: List[Bounds]
    ) -> bool:
        """
        Tracks the camera movements according to the changing background
        inside the ROI.

        Parameters
        ----------
        prev_frame, frame : np.ndarray
            Frames used to detect background movement.
        igonerd_regions : List[Bounds]
            Tracked object's boundaries.

            Tracked object's does not form part of the background so
            they should be ignored.

        Returns
        -------
        bool
            Whether or not good points were found or sucessfully
            tracked.
        """

        # Initialize a mask of what to track
        height, weight = frame.shape[:2]
        mask = 255 * np.ones((height, weight), dtype=np.uint8)

        # Mask pixeles inside every ROIs
        for x_0, x_f, y_0, y_f in ignored_regions:
            mask[y_0:y_f, x_0:x_f] = 0

        p_good, affine_params, err = _get_affine(
            img1=prev_frame, img2=frame, region=self.roi._get_bounds(), mask=mask
        )
        self.features = p_good[1:]

        if err is None:
            return False

        self.affine_params_history.append(affine_params)
        self.mse.append(err)

        return True


class TrackingScenario:
    """
    Controls all the tracking process along the video.

    Parameters
    ----------
    object_trackers : list
        Trackers of all the objects.
    camera_tracker : CameraTracker
        Tracker used to detect camera movements, by default None.
    undistorter : Undistorter
        Undistorted used to correct each video frame, by default None.
    preview_scale : float
        Scale of the video preview, by default 1.0.
    auto_mode : bool
        If True the video is processed auomtically otherwise it's
        processed manually, by default True.

        If the video is processed manually, pressing ``ENTER`` key is
        necessary in every frame to continue.

        This mode can be changed in the middle of the processing by
        pressing ``M`` key.

    Attributes
    ----------
    object_trackers : list
        Trackers of all the objects
    camera_tracker : CameraTracker
        Tracker used to detect camera movements.
    undistorter : Undistorter
        Undistorted used to correct each video frame.
    preview_scale : float
        Scale of the video preview.
    auto_mode : bool
        If True the video is processed auomtically otherwise it's
        processed manually, by default True.

        If the video is processed manually, pressing ``Enter`` key is
        necessary in every frame to continue.

        This mode can be changed in the middle of the processing by
        pressing ``M`` key.
    """

    def __init__(
        self,
        object_trackers: List[ObjectTracker],
        camera_tracker: Optional[CameraTracker] = None,
        undistorter: Optional[Undistorter] = None,
        preview_scale: float = 1,
        auto_mode: bool = True,
    ):
        self.object_trackers = object_trackers
        self.camera_tracker = camera_tracker
        self.undistorter = undistorter
        self.preview_scale = preview_scale
        self.auto_mode = auto_mode
        self._enabled = True
        self._iteration_counter = 0

        self.video_path: str
        self.cap: Any
        self.frame_count: int
        self.fps: int
        self.width: int
        self.height: int
        self.dim: Tuple[int, int]
        self.prev_frame: np.ndarray
        self.first_frame: int
        self.last_frame: Optional[int] = None

    def _digest_video_path(self, video_path):
        if not Path.exists(Path(video_path)):
            raise ValueError(f"Path '{video_path}' does not exists")
        self.video_path = video_path

        # Create capture object
        self.cap = cv2.VideoCapture(video_path)

        # Total number of frames in the video file
        self.frame_count = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))

        # Frames per seconds
        self.fps = self.cap.get(cv2.CAP_PROP_FPS)

        # Frame width
        self.width = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))

        # Frame height
        self.height = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

        self.dim = (self.width, self.height)

        self.first_frame = 0

    def _undistort(self, frame):
        if self.undistorter:
            frame = self.undistorter.fix(frame)
        return frame

    def _show_frame(self, frame, show_frame_id=True):
        # CXY, region, features, frame_numb, mask
        frame = frame.copy()

        # Draw region in which features are detected
        if self.camera_tracker:
            x_0, x_f, y_0, y_f = self.camera_tracker.roi._get_bounds()

            cv2.putText(
                img=frame,
                text="Camera Tracking region",
                org=(x_0 + 5, y_f - 5),
                fontFace=cv2.FONT_HERSHEY_COMPLEX_SMALL,
                fontScale=1.2,
                color=(0, 0, 255),
                thickness=1,
                lineType=cv2.LINE_AA,
            )

            cv2.rectangle(frame, (x_0, y_0), (x_f, y_f), (0, 0, 255), 2)
            p_2, p_3 = self.camera_tracker.features
            # Draw detected and estimated features
            for p2_, p3_ in zip(p_2, p_3):
                x_2, y_2 = np.rint(p2_).astype(np.int32)
                x_3, y_3 = np.rint(p3_).astype(np.int32)

                cv2.circle(frame, (x_2, y_2), 3, (0, 0, 0), -1)
                cv2.circle(frame, (x_3, y_3), 3, (0, 255, 0), -1)

        for obj_tracker in self.object_trackers:
            # TODO: Do this better:
            # Alter the blue channel in ant-related pixels
            window = obj_tracker.roi._crop(frame, prev=True)
            if obj_tracker.mask is not None:
                window[:, :, 0] = obj_tracker.mask

            # Draw a point over the roi center and draw bounds
            x_1, x_2, y_1, y_2 = obj_tracker.roi._get_bounds()
            cv2.circle(frame, obj_tracker.roi._centroid, 5, (255, 255, 255), -1)
            cv2.rectangle(frame, (x_1, y_1), (x_2, y_2), (0, 255, 255), 2)
            cv2.putText(
                img=frame,
                text=obj_tracker.name,
                org=(x_1 + 5, y_2 - 5),
                fontFace=cv2.FONT_HERSHEY_COMPLEX_SMALL,
                fontScale=1.2,
                color=(0, 255, 255),
                thickness=1,
                lineType=cv2.LINE_AA,
            )

        if show_frame_id:
            height, weight = frame.shape[:2]
            frame_id = self._iteration_counter + self.first_frame
            x, y = int(0.02 * weight), int(0.05 * height)
            cv2.putText(
                img=frame,
                text=str(frame_id),
                org=(x, y),
                fontFace=cv2.FONT_HERSHEY_COMPLEX_SMALL,
                fontScale=1.2,
                color=(0, 255, 255),
                thickness=1,
                lineType=cv2.LINE_AA,
            )

        frame = _resize_frame(frame, self.preview_scale)
        cv2.imshow("yupi processing window", frame)

    def _create_ui(
        self,
        img: np.ndarray,
        t_name: str,
        current_tracker: int,
        total_trackers: int,
        roi: ROI,
    ):
        imgc = img.copy()
        imgc = _resize_frame(imgc, roi.scale)
        height = img.shape[0] * roi.scale
        weight = img.shape[1] * roi.scale
        h_pad = 0.2
        w_pad = 0.15

        imgc = cv2.blur(imgc, (5, 5))

        box = imgc[
            int(height * h_pad) : int(height - height * h_pad),
            int(weight * w_pad) : int(weight - weight * w_pad),
            :,
        ]
        threshold = 180

        box[:, :] += np.array([threshold, threshold, threshold], dtype="uint8")
        box[box < threshold] = 255

        bshape = box.shape
        boxw = bshape[1]
        boxh = bshape[0]

        # Text
        font = cv2.FONT_HERSHEY_SIMPLEX
        font_scale = 0.0008333 * weight + 0.0333333
        color = (50, 50, 50)
        thickness = int(font_scale + 1)

        text_lines = [
            "Your tracking scenario is almost ready",
            "Let's initialize your trackers",
            "Next, you will have to click on the initial",
            f"position of the tracker {t_name.upper()}",
            "Press any key to continue...",
            f"Trackers Initialized: {current_tracker}/{total_trackers}",
        ]

        l = int(0.0396825 * boxw + 2.222222)

        def put_text(img: np.ndarray, text: str, pos: Tuple[float, float]):
            return cv2.putText(
                img, text, pos, font, font_scale, color, thickness, cv2.LINE_AA
            )

        box = put_text(box, text_lines[0], (l, l + l))
        box = put_text(box, text_lines[1], (l, l + 2 * l))
        box = put_text(box, text_lines[2], (l, l + 4 * l))
        box = put_text(box, text_lines[3], (l, l + 5 * l))
        box = put_text(box, text_lines[4], (l, l + 7 * l))
        box = put_text(box, text_lines[5], (l, boxh - l))
        return imgc

    def _first_iteration(self, start_frame):
        # Start processing frams at the given index
        if start_frame:
            self.first_frame = start_frame
            self.cap.set(cv2.CAP_PROP_POS_FRAMES, start_frame)

        # Capture the first frame to process
        _, prev_frame = self.cap.read()

        # Correct spherical distortion
        self.prev_frame = self._undistort(prev_frame)

        # Initialize the roi of all the trackers
        for i, obj_tracker in enumerate(self.object_trackers):
            if obj_tracker.roi.init_mode == "manual":
                tracker_name = obj_tracker.name
                ui = self._create_ui(
                    self.prev_frame,
                    tracker_name,
                    i,
                    len(self.object_trackers),
                    obj_tracker.roi,
                )
                cv2.imshow(
                    "Initialization of trackers: Press any key to start with "
                    f"tracker: {tracker_name.upper()}",
                    ui,
                )
                cv2.waitKey(-1)
                cv2.destroyAllWindows()
            retval = obj_tracker._init_roi(self.prev_frame)
            if not retval:
                return retval

        # Initialize the region of the camera tracker
        if self.camera_tracker:
            self.camera_tracker._init_roi(self.prev_frame)

        # Increase the iteration counter
        self._iteration_counter += 1

        logging.info("All trackers were initialized")
        return True

    def _keyboard_controller(self):
        # Keyboard events
        wait_key = 0 if not self.auto_mode else 10

        key = cv2.waitKey(wait_key) & 0xFF
        if key == ord("m"):
            self.auto_mode = not self.auto_mode

        elif key == ord("q"):
            self._enabled = False

        elif key == ord("e"):
            exit()

    def _regular_iteration(self):
        # Get current frame and ends the processing when no more frames are
        # detected

        frame_id = self._iteration_counter + self.first_frame
        if self.last_frame is not None and frame_id >= self.last_frame:
            return False, True

        ret, frame = self.cap.read()
        if not ret:
            logging.info("All frames were processed")
            return False, True

        # Correct spherical distortion
        frame = self._undistort(frame)

        # ROI Arrays of tracking objects
        roi_array = []

        # Track every object and save past and current ROIs
        for otrack in self.object_trackers:
            roi_array.append(otrack.roi._get_bounds())
            otrack._track(frame)
            roi_array.append(otrack.roi._get_bounds())

        if self.camera_tracker:
            ret = self.camera_tracker._track(self.prev_frame, frame, roi_array)

        if not ret:
            msg = f"CameraTracker - No matrix was estimated (Frame {frame_id})"
            logging.error(msg)
            return False, False

        # Display the full image with the ant in blue (TODO: Refactor this to
        # Make more general)
        self._show_frame(frame)

        for otrack in self.object_trackers:
            otrack.roi._prev_centroid = otrack.roi._centroid

        # Save current frame and ROI center as previous for next iteration
        self.prev_frame = frame.copy()

        # Call the keyboard controller to handle key interruptions
        self._keyboard_controller()

        # Increase the iteration counter
        self._iteration_counter += 1

        return True, False

    def _release_cap(self):
        self.cap.release()
        cv2.destroyAllWindows()

    def _tracker2trajectory(self, tracker, pix_per_m):
        dt = 1 / self.fps
        traj_id = tracker.name
        x, y = map(list, zip(*tracker.history))
        x_arr = np.array(x) / pix_per_m
        y_arr = -1 * np.array(y) / pix_per_m
        return Trajectory(x=x_arr, y=y_arr, dt=dt, traj_id=traj_id)

    def _export_trajectories(self, pix_per_m):
        t_list = []
        reference = None
        # Extract camera reference
        if self.camera_tracker:
            affine_params = np.array(self.camera_tracker.affine_params_history)
            theta, t_x, t_y, _ = affine_params.T
            t_x, t_y = t_x / pix_per_m, t_y / pix_per_m
            # Invert axis
            theta *= -1
            t_y *= -1
            reference = theta, t_x, t_y
        # Output the trajectory of each tracker
        for otrack in self.object_trackers:
            t = self._tracker2trajectory(otrack, pix_per_m)
            if self.camera_tracker:
                assert reference is not None
                t = add_moving_FoR(t, reference, new_traj_id=t.traj_id)
            t_list.append(t)
        return t_list

    def track(
        self,
        video_path: str,
        start_frame: int = 0,
        end_frame: Optional[int] = None,
        pix_per_m: int = 1,
    ) -> Tuple[bool, Optional[List[Trajectory]]]:
        """
        Starts the tracking process.

        Parameters
        ----------
        video_path : str
            Path of the video used to track the objects.
        start_frame : int, optional
            Initial frame in which starts the processing, by default 0.
        end_frame : Optional[int]
            Last frame being processed, if nothing is passed all frames
            until the end of the video will be processed, by default
            None.
        pix_per_m : int, optional
            Pixel per meters, by default 1.

            This value is used to readjuts the trajectories points to a
            real scale.

        Returns
        -------
        bool
            Whether or not the tracking process ended succefully.
        List[Trajectory]
            List of all the trajectories extracted in the tracking
            process.
        """

        if end_frame is not None and end_frame > start_frame:
            self.last_frame = int(end_frame)

        self._digest_video_path(video_path)

        end = False
        retval = False
        if self._iteration_counter == 0:
            retval = self._first_iteration(start_frame)
            if not retval:
                return retval, None

        logging.info("Processing frames")
        while self._enabled:
            retval, end = self._regular_iteration()
            if not retval:
                break

        if end:
            retval = True

        self._release_cap()
        trajectories = self._export_trajectories(pix_per_m)
        return retval, trajectories



"""
This contains spatial plotting functions for the trajectories.
"""

import itertools
import logging
import warnings
from typing import Callable, Collection, List, Optional, Union

import matplotlib.pyplot as plt

from yupi import Trajectory
from yupi.graphics._style import LINE, YUPI_COLORS


def plot_2d(
    trajs: Union[List[Trajectory], Trajectory],
    line_style: str = LINE,
    title: Optional[str] = None,
    legend: bool = True,
    show: bool = True,
    connected: bool = False,
    units: str = "m",
    color=None,
    **kwargs,
):
    """
    Plot all the points of trajectories from ``trajs`` in a 2D plane.

    Parameters
    ----------
    trajs : Union[List[Trajectory], Trajectory]
        Input trajectories.
    line_style : str
        Type of the trajectory line to plot. It uses the matplotlib,
        notation, by default '-'.
    title : str, optional
        Title of the plot, by default None.
    legend : bool, optional
        If True, legend is shown. By default True.
    show : bool, optional
        If True, the plot is shown. By default True.
    connected : bool
        If True, all the trajectory points of same index will be,
        connected.

        If the trajectories do not have same length then the points
        will be connected until the shortest trajectory last index.
    color : str or tuple or list
        Defines the color of the trajectories, by default None.

        If color is of type ``str`` or ``tuple`` (rgb) then the color
        is applied to all trajectories. If color is of type ``list``
        then the trajectories take the color according to the index.

        If there are less colors than trajectories then the remaining
        trajectories are colored automatically (not with the same
        color).
    """

    if isinstance(trajs, Trajectory):
        trajs = [trajs]

    units = "" if units is None else f" [{units}]"
    cycle = itertools.cycle(YUPI_COLORS)
    colors = [next(cycle) for _ in trajs]
    ax = plt.gca()

    if color is not None:
        if isinstance(color, (str, tuple)):
            kwargs["color"] = color
        elif isinstance(color, list):
            colors = color

    if connected:
        lengths = list(map(len, trajs))
        min_len = min(lengths)
        max_len = max(lengths)
        if min_len != max_len:
            logging.warning("Not all the trajectories have the same length.")
        for i in range(min_len):
            traj_points = [t[i] for t in trajs]
            traj_points.append(traj_points[0])
            for tp1, tp2 in zip(traj_points[:-1], traj_points[1:]):
                seg_x = [tp1.r[0], tp2.r[0]]
                seg_y = [tp1.r[1], tp2.r[1]]
                plt.plot(seg_x, seg_y, color=(0.2, 0.2, 0.2), linewidth=0.5)

    for i, traj in enumerate(trajs):

        if traj.dim != 2:
            logging.warning(
                "Using plot_2d with a trajectory of %i dimensions"
                " Trajectory No. %i with id %s",
                traj.dim,
                i,
                traj.traj_id,
            )

        # Plotting
        x_data, y_data = traj.r.x, traj.r.y
        if colors is not None:
            if i < len(colors):
                kwargs["color"] = colors[i]
            else:
                kwargs.pop("color")
        traj_plot = plt.plot(x_data, y_data, line_style, **kwargs)
        color = traj_plot[-1].get_color()
        traj_id = traj.traj_id if traj.traj_id else f"traj {i}"
        plt.plot(
            x_data[0],
            y_data[0],
            "o",
            mfc="white",
            zorder=2,
            label=f"{traj_id} start",
            color=color,
        )
        plt.plot(x_data[-1], y_data[-1], "o", mfc="white", zorder=2, color=color)
        plt.plot(
            x_data[-1],
            y_data[-1],
            "o",
            alpha=0.5,
            label=f"{traj_id} end",
            color=color,
        )

        if legend:
            plt.legend()

        plt.title(title)
        plt.tick_params(direction="in")
        plt.axis("equal")
        plt.grid(True)
        plt.xlabel(f"x{units}")
        plt.ylabel(f"y{units}")

    if show:
        plt.show()

    return ax


def plot_2D(  # pylint: disable=invalid-name
    trajs: Union[List[Trajectory], Trajectory],
    line_style: str = LINE,
    title: Optional[str] = None,
    legend: bool = True,
    show: bool = True,
    connected: bool = False,
    units: str = "m",
    color=None,
    **kwargs,
):
    """
    .. deprecated:: 0.10.0
        :func:`plot_2D` will be removed in a future version, use
        :func:`plot_2d` instead.

    Plot all the points of trajectories from ``trajs`` in a 2D plane.

    Parameters
    ----------
    trajs : Union[List[Trajectory], Trajectory]
        Input trajectories.
    line_style : str
        Type of the trajectory line to plot. It uses the matplotlib,
        notation, by default '-'.
    title : str, optional
        Title of the plot, by default None.
    legend : bool, optional
        If True, legend is shown. By default True.
    show : bool, optional
        If True, the plot is shown. By default True.
    connected : bool
        If True, all the trajectory points of same index will be,
        connected.

        If the trajectories do not have same length then the points
        will be connected until the shortest trajectory last index.
    color : str or tuple or list
        Defines the color of the trajectories, by default None.

        If color is of type ``str`` or ``tuple`` (rgb) then the color
        is applied to all trajectories. If color is of type ``list``
        then the trajectories take the color according to the index.

        If there are less colors than trajectories then the remaining
        trajectories are colored automatically (not with the same
        color).
    """

    warnings.warn(
        "plot_2D is deprecated and will be removed in a future version, "
        "use plot_2d instead",
        DeprecationWarning,
    )
    plot_2d(
        trajs,
        line_style,
        title,
        legend,
        show,
        connected,
        units,
        color,
        **kwargs,
    )


def plot_3d(
    trajs: Union[List[Trajectory], Trajectory],
    line_style: str = LINE,
    title: Optional[str] = None,
    legend: bool = True,
    show: bool = True,
    connected: bool = False,
    units: str = "m",
    color=None,
    **kwargs,
):
    """
    Plot all the points of trajectories from ``trajs`` in a 3D space.

    Parameters
    ----------
    trajs : Union[List[Trajectory], Trajectory]
        Input trajectories.
    line_style : str
        Type of the trajectory line to plot. It uses the matplotlib,
        notation, by default '-'.
    title : str, optional
        Title of the plot, by default None.
    legend : bool, optional
        If True, legend is shown. By default True.
    show : bool, optional
        If True, the plot is shown. By default True.
    connected : bool
        If True, all the trajectory points of same index will be,
        connected.

        If the trajectories do not have same length then the points
        will be connected until the shortest trajectory last index.
    color : str or tuple or list
        Defines the color of the trajectories, by default None.

        If color is of type ``str`` or ``tuple`` (rgb) then the color
        is applied to all trajectories. If color is of type ``list``
        then the trajectories take the color according to the index.

        If there are less colors than trajectories then the remaining
        trajectories are colored automatically (not with the same
        color).
    """

    if isinstance(trajs, Trajectory):
        trajs = [trajs]

    units = "" if units is None else f" [{units}]"

    cycle = itertools.cycle(YUPI_COLORS)
    colors = [next(cycle) for _ in trajs]

    if color is not None:
        if isinstance(color, (str, tuple)):
            kwargs["color"] = color
        elif isinstance(color, list):
            colors = color

    ax = plt.axes(projection="3d")

    if connected:
        lengths = list(map(len, trajs))
        min_len = min(lengths)
        max_len = max(lengths)
        if min_len != max_len:
            logging.warning("Not all the trajectories have the same length.")
        for i in range(min_len):
            traj_points = [t[i] for t in trajs]
            traj_points.append(traj_points[0])
            for tp1, tp2 in zip(traj_points[:-1], traj_points[1:]):
                seg_x = [tp1.r[0], tp2.r[0]]
                seg_y = [tp1.r[1], tp2.r[1]]
                seg_z = [tp1.r[2], tp2.r[2]]
                ax.plot(seg_x, seg_y, seg_z, color=(0.2, 0.2, 0.2), linewidth=0.5)

    for i, traj in enumerate(trajs):

        if traj.dim != 3:
            logging.warning(
                "Using plot_3d with a trajectory of %i dimensions"
                " Trajectory No. %i with id %s",
                traj.dim,
                i,
                traj.traj_id,
            )

        # Plotting
        x_data, y_data, z_data = traj.r.x, traj.r.y, traj.r.z

        if colors is not None:
            if i < len(colors):
                kwargs["color"] = colors[i]
            else:
                kwargs.pop("color")
        traj_plot = ax.plot(x_data, y_data, z_data, line_style, **kwargs)
        color = traj_plot[-1].get_color()
        traj_id = traj.traj_id if traj.traj_id else f"traj {i}"

        ax.plot(
            x_data[0],
            y_data[0],
            z_data[0],
            "o",
            mfc="white",
            label=f"{traj_id} start",
            color=color,
        )

        ax.plot(x_data[-1], y_data[-1], z_data[-1], "o", mfc="white", color=color)
        ax.plot(
            x_data[-1],
            y_data[-1],
            z_data[-1],
            "o",
            alpha=0.5,
            label=f"{traj_id} end",
            color=color,
        )

        if legend:
            plt.legend()

        plt.title(title)
        plt.tick_params(direction="in")
        plt.grid(True)
        ax.set_xlabel(f"x{units}")
        ax.set_ylabel(f"y{units}")
        ax.set_zlabel(f"z{units}")

    if show:
        plt.show()

    return ax


def plot_3D(  # pylint: disable=invalid-name
    trajs: Union[List[Trajectory], Trajectory],
    line_style: str = LINE,
    title: Optional[str] = None,
    legend: bool = True,
    show: bool = True,
    connected: bool = False,
    units: str = "m",
    color=None,
    **kwargs,
):
    """
    .. deprecated:: 0.10.0
        :func:`plot_3D` will be removed in a future version, use
        :func:`plot_3d` instead.

    Plot all the points of trajectories from ``trajs`` in a 3D space.

    Parameters
    ----------
    trajs : Union[List[Trajectory], Trajectory]
        Input trajectories.
    line_style : str
        Type of the trajectory line to plot. It uses the matplotlib,
        notation, by default '-'.
    title : str, optional
        Title of the plot, by default None.
    legend : bool, optional
        If True, legend is shown. By default True.
    show : bool, optional
        If True, the plot is shown. By default True.
    connected : bool
        If True, all the trajectory points of same index will be,
        connected.

        If the trajectories do not have same length then the points
        will be connected until the shortest trajectory last index.
    color : str or tuple or list
        Defines the color of the trajectories, by default None.

        If color is of type ``str`` or ``tuple`` (rgb) then the color
        is applied to all trajectories. If color is of type ``list``
        then the trajectories take the color according to the index.

        If there are less colors than trajectories then the remaining
        trajectories are colored automatically (not with the same
        color).
    """
    warnings.warn(
        "plot_3D is deprecated and will be removed in a future version, "
        "use plot_3d instead",
        DeprecationWarning,
    )
    plot_3d(
        trajs,
        line_style,
        title,
        legend,
        show,
        connected,
        units,
        color,
        **kwargs,
    )


def plot_vs_time(
    trajs: Union[List[Trajectory], Trajectory],
    key: Callable[[Trajectory], Collection[float]],
    line_style: str = LINE,
    x_units: str = "s",
    y_label: Union[str, None] = None,
    title: Optional[str] = None,
    legend: bool = True,
    color = None,
    show: bool = True,
    **kwargs,
):
    if isinstance(trajs, Trajectory):
        trajs = [trajs]

    x_units = "time" + ("" if x_units is None else f" [{x_units}]")

    cycle = itertools.cycle(YUPI_COLORS)
    colors = [next(cycle) for _ in trajs]

    if color is not None:
        if isinstance(color, (str, tuple)):
            kwargs["color"] = color
        elif isinstance(color, list):
            colors = color

    for i, traj in enumerate(trajs):
        if colors is not None:
            if i < len(colors):
                kwargs["color"] = colors[i]
            else:
                kwargs.pop("color")
        y_data = key(traj)
        x_data = traj.t
        traj_id = traj.traj_id if traj.traj_id else f"traj {i}"
        plt.plot(x_data, y_data, line_style, **kwargs, label=traj_id)
        plt.xlabel(x_units)
        if y_label is not None:
            plt.ylabel(y_label)
        plt.grid()
        plt.title(title)

    if legend:
        plt.legend()

    if show:
        plt.show()

    return plt.gca()



from typing import Optional, Tuple

import numpy as np

from yupi import Trajectory
from yupi.transformations._affine_estimator import _affine_matrix


def _affine2camera(
    theta: np.ndarray, t_x: np.ndarray, t_y: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

    x_cam2lab, y_cam2lab, theta_cam2lab = np.zeros((3, theta.shape[0] + 1))
    theta_cam2lab[1:] = np.cumsum(theta)

    for i in range(theta.size):
        affine_mat = _affine_matrix(
            theta_cam2lab[i + 1], x_cam2lab[i], y_cam2lab[i], inverse=True
        )
        shift_vec = [-t_x[i], -t_y[i], 1]
        x_cam2lab[i + 1], y_cam2lab[i + 1] = affine_mat @ shift_vec

    return x_cam2lab[1:], y_cam2lab[1:], theta_cam2lab[1:]


def _camera2obj(
    x_obj2cam: np.ndarray,
    y_obj2cam: np.ndarray,
    x_cam2lab: np.ndarray,
    y_cam2lab: np.ndarray,
    theta_cam2lab: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:

    x_obj2lab, y_obj2lab = np.empty((2, x_obj2cam.size))

    for i in range(x_obj2cam.size):
        affine_mat = _affine_matrix(
            theta_cam2lab[i], x_cam2lab[i], y_cam2lab[i], inverse=True
        )
        shift_vec = [x_obj2cam[i], y_obj2cam[i], 1]
        x_obj2lab[i], y_obj2lab[i] = affine_mat @ shift_vec

    return x_obj2lab, y_obj2lab


def _affine2obj(
    theta: np.ndarray,
    t_x: np.ndarray,
    t_y: np.ndarray,
    x_obj2cam: np.ndarray,
    y_obj2cam: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:

    x_cam2lab, y_cam2lab, theta_cam2lab = _affine2camera(theta, t_x, t_y)
    x_obj2lab, y_obj2lab = _camera2obj(
        x_obj2cam, y_obj2cam, x_cam2lab, y_cam2lab, theta_cam2lab
    )
    return x_obj2lab, y_obj2lab


def add_moving_FoR(  # pylint: disable=invalid-name
    traj: Trajectory,
    reference: Tuple[np.ndarray, np.ndarray, np.ndarray],
    start_at_origin: bool = True,
    new_traj_id: Optional[str] = None,
) -> Trajectory:
    """
    This function fuses the information of a trajectory with an
    external reference of the motion of the Frame of Reference
    (FoR).

    It allows to remap the information gathered in local SoRs
    to a more general FoR.

    Parameters
    ----------
    traj : Trajectory
        Input trajectory.
    reference : Tuple[np.ndarray,np.ndarray,np.ndarray]
        Angular and translational parameters of the form
        ``(ang:np.ndarray, tx:np.ndarray, ty:np.ndarray)`` that
        accounts for the orientation and displacement of the reference.
    start_at_origin : bool, optional
        If True, set initial position at the origin. By default True.

    Returns
    -------
    Trajectory
        Output trajectory in the lab frame of reference.
    """

    theta, t_x, t_y = reference
    x_al, y_al = _affine2obj(theta, t_x, t_y, traj.r.x, traj.r.y)

    if start_at_origin:
        x_al = x_al - x_al[0]
        y_al = y_al - y_al[0]

    moved_traj = Trajectory(
        x=x_al,
        y=y_al,
        t=traj.t,
        traj_id=new_traj_id,
        diff_est=traj.diff_est,
    )
    return moved_traj



"""
Contains the basic structures for trajectories.
"""

from __future__ import annotations

import csv
import json
import logging
import os
import warnings
from pathlib import Path
from typing import (
    Any,
    Collection,
    Dict,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
    cast,
)

import numpy as np

import yupi._differentiation as diff
from yupi.exceptions import LoadTrajectoryError
from yupi.features import Features
from yupi.vector import Vector

_THRESHOLD = 1e-12

Axis = Collection[float]
"""Represents the data for a single axis."""

Point = Collection[float]
"""Represents a single point."""


class TrajectoryPoint(NamedTuple):
    """
    Represents a point of a trajectory.

    Parameters
    ----------
    r : Vector
        Positional data.
    v : Vector
        Velocity data.
    t : float
        Time data.
    """

    r: Vector
    v: Vector
    t: float


class Trajectory:
    """
    A Trajectory object represents a multidimensional trajectory.
    It can be iterated to obtain the corresponding point for each
    timestep.

    Parameters
    ----------
    x : Optional[Axis]
        Array containing position data of X axis, by default None
    y : Optional[Axis]
        Array containing position data of Y axis, by default None.
    z : Optional[Axis]
        Array containing position data of X axis, by default None.
    points : Optional[Collection[Point]]
        Array containing position data as a list of points, by default
        None
    axes : Optional[Collection[Axis]]
        Array containing position data as a list of axis, by default
        None
    t : Optional[Collection[float]]
        Array containing time data, by default None.
    dt : float
        If no time data is given this represents the time between each
        position data value.
    t_0 : float
        If no time data is given this represents the initial time value,
        by default 0.
    traj_id : str
        Id of the trajectory.
    lazy : bool
        Defines if the velocity vector is not recalculated every time
        is asked. By default False.
    diff_est : Dict[str, Any]
        Dictionary containing the parameters for the differentiation
        estimation method used to calculate velocity.

    Attributes
    ----------
    r : Vector
        Position vector.
    dt_mean : float
        Mean of the time data delta.
    dt_std : float
        Standard deviation of the time between each position data value.
    traj_id : str
        Id of the trajectory.
    lazy : bool
        Defines if the velocity vector is not recalculated every time
        is asked.
    diff_est : dict
        Dictionary containing the parameters for the differentiation
        estimation method used to calculate velocity.

    Examples
    --------
    You can create a trajectory object by giving the arrays that
    represent it:

    >>> x = [0, 1.2, 3, 2.8]
    >>> y = [0, 3.1, 0.7, 1.6]
    >>> Trajectory(x=x, y=y)

    You can also create the trajectory given the points:

    >>> points = [[0, 0], [1.2, 3.1], [3, 0.7], [2.8, 1.6]]
    >>> Trajectory(points=points)

    Or even create it given all the data for each dimension in a single
    source:

    >>> axes = [[0, 1.2, 3, 2.8], [0, 3.1, 0.7, 1.6]]
    >>> Trajectory(axes=axes)

    All of these examples create the same trajectory.

    Raises
    ------
    ValueError
        If positional data is given in more than one way.
    ValueError
        If no positional data is given.
    ValueError
        If all the given input data (``x``, ``y``, ``z``, ``t``)
        does not have the same length.
    ValueError
        If ``t`` and ``dt`` given but ``t`` is not uniformly spaced.
    ValueError
        If ``t`` and ``dt`` given but ``dt`` does not match ``t``
        values delta.
    """

    general_diff_est: Dict[str, Any] = {
        "method": diff.DiffMethod.LINEAR_DIFF,
        "window_type": diff.WindowType.CENTRAL,
    }

    def __init__(
        self,
        x: Optional[Axis] = None,
        y: Optional[Axis] = None,
        z: Optional[Axis] = None,
        points: Optional[Collection[Point]] = None,
        axes: Optional[Collection[Axis]] = None,
        t: Optional[Collection[float]] = None,
        dt: Optional[float] = None,
        t_0: float = 0.0,
        traj_id: str = "",
        lazy: Optional[bool] = False,
        diff_est: Optional[Dict[str, Any]] = None,
        vel_est: Optional[Dict[str, Any]] = None,
        t0: Optional[float] = None,  # pylint: disable=invalid-name
    ):  # pylint: disable=too-many-arguments

        # Position data validation
        from_xyz = x is not None
        from_points = points is not None
        from_axes = axes is not None

        if from_xyz + from_points + from_axes > 1:
            raise ValueError(
                "Positional data must come only from one way: "
                "'xyz' data, 'points' data or 'axes' data."
            )

        # Set position data
        lengths = [len(t)] if t is not None else []

        # xyz data is converted to axes
        if from_xyz:
            axes = [d for d in [x, y, z] if d is not None]

        # Check if positional data is given
        if axes is not None and len(axes) > 0:
            lengths.extend([len(d) for d in axes])
            self.r = Vector(axes, dtype=float, copy=True).T
        elif points is not None:
            lengths.append(len(points))
            self.r = Vector(points, dtype=float, copy=True)
        else:
            raise ValueError("No position data were given.")

        # Check if all the given data has the same shape
        if lengths.count(lengths[0]) != len(lengths):
            raise ValueError("All input arrays must have the same lenght.")
        if len(self.r) < 2:
            raise ValueError("The trajectory must contain at least 2 points.")

        if t0 is not None:
            t_0 = t0
            warnings.warn(
                "'t0' is deprecated and will be removed in a future version, "
                "use 't_0' instead.",
                DeprecationWarning,
            )

        self.__dt = dt
        self.t_0 = t_0
        self.__t = None if t is None else Vector(t, dtype=float, copy=True)
        self.__v: Optional[Vector] = None
        self.__a: Optional[Vector] = None
        self.traj_id = traj_id
        self.lazy = lazy

        # Set time data
        if self.__t is None:
            self.dt_mean = dt if dt is not None else 1.0
            self.dt_std = 0
        else:
            self.dt_mean = np.mean(np.array(self.__t.delta))
            self.dt_std = np.std(np.array(self.__t.delta))

        # Differentiation method
        if vel_est is not None:
            diff_est = vel_est
            warnings.warn(
                "'vel_est' is deprecated and will be removed in a future version, "
                "use 'diff_est' instead.",
                DeprecationWarning,
            )
        self.diff_est = Trajectory.general_diff_est.copy()
        if diff_est is not None:
            self.diff_est.update(diff_est)

        # Time parameters validation
        if self.__t is not None and dt is not None:
            if abs(self.dt_mean - dt) > _THRESHOLD:
                raise ValueError(
                    "You are giving 'dt' and 't' but 'dt' "
                    "does not match with time values delta."
                )
            if abs(self.dt_std - 0) > _THRESHOLD:
                raise ValueError(
                    "You are giving 'dt' and 't' but 't' is not uniformly spaced."
                )
            if abs(self.__t[0] - t_0) > _THRESHOLD:
                raise ValueError(
                    "You are giving 'dt' and 't' but 't_0' is not "
                    "the same as the first value of 't'."
                )

        self.features = Features(self)

    def set_diff_method(
        self,
        method: diff.DiffMethod,
        window_type: diff.WindowType = diff.WindowType.CENTRAL,
        accuracy: int = 1,
    ):
        """
        Set the local diferentiation method.

        Parameters
        ----------
        method : DiffMethod
            Method used to differentiate.
        window_type : WindowType
            Type of window used in the differentiation method. By default,
            the central window is used.
        accuracy : int
            Accuracy of the differentiation method (only valid for
            FORNBERG_DIFF method). By default, the accuracy is 1.
        """
        self.diff_est = {
            "method": method,
            "window_type": window_type,
            "accuracy": accuracy,
        }
        self.recalculate_velocity()

    def set_vel_method(
        self,
        method: diff.DiffMethod,
        window_type: diff.WindowType = diff.WindowType.CENTRAL,
        accuracy: int = 1,
    ):
        """
        .. deprecated:: 0.10.0
            :func:`set_vel_method` is deprecated and will be removed in
            version 1.0.0, use :func:`set_diff_method` instead.
        """
        warnings.warn(
            "'set_vel_method' is deprecated and will be removed in a future version, "
            "use 'set_diff_method' instead.",
            DeprecationWarning,
        )
        self.set_diff_method(method, window_type, accuracy)

    @staticmethod
    def global_diff_method(
        method: diff.DiffMethod,
        window_type: diff.WindowType = diff.WindowType.CENTRAL,
        accuracy: int = 1,
    ):
        """
        Set the global diferentiation method.

        Parameters
        ----------
        method : DiffMethod
            Method used to differentiate.
        window_type : WindowType
            Type of window used in the differentiation method. By default,
            the central window is used.
        accuracy : int
            Accuracy of the differentiation method (only valid for
            FORNBERG_DIFF method). By default, the accuracy is 1.
        """
        Trajectory.general_diff_est = {
            "method": method,
            "window_type": window_type,
            "accuracy": accuracy,
        }

    @staticmethod
    def global_vel_method(
        method: diff.DiffMethod,
        window_type: diff.WindowType = diff.WindowType.CENTRAL,
        accuracy: int = 1,
    ):
        """
        .. deprecated:: 0.10.0
            :func:`global_vel_method` is deprecated and will be removed in
            version 1.0.0, use :func:`global_diff_method` instead.
        """
        warnings.warn(
            "'global_vel_method' is deprecated and will be removed in "
            "version 1.0.0, use 'global_diff_method' instead.",
            DeprecationWarning,
        )
        Trajectory.global_diff_method(method, window_type, accuracy)

    @property
    def dt(self) -> float:
        """
        Returns the time between each position data value.

        If the time data is not uniformly spaced it returns an
        estimated value.
        """
        return self.dt_mean if self.__dt is None else self.__dt

    @property
    def uniformly_spaced(self) -> bool:
        """bool : True if the time data is uniformly spaced"""
        if self.__t is not None:
            starts_at_zero = self.__t[0] == 0
            std_is_zero = self.dt_std == 0
            return starts_at_zero and std_is_zero
        return True

    def __len__(self) -> int:
        return self.r.shape[0]

    def __getitem__(self, index) -> Union[Trajectory, TrajectoryPoint]:
        if isinstance(index, int):
            # r, v, t
            data = [self.r[index], None, None]
            data[1] = self.v[index - 1] if index > 0 else Vector([0] * self.dim)
            data[2] = (
                self.t[index] if self.__t is not None else self.t_0 + index * self.dt
            )

            r, v, t = data
            return TrajectoryPoint(r=r, v=v, t=t)

        if isinstance(index, slice):
            start, stop, step = index.indices(len(self))
            new_points = self.r[start:stop:step]
            if self.uniformly_spaced:
                new_dt = self.dt * step
                new_t0 = self.t_0 + start * self.dt
                return Trajectory(
                    points=new_points,
                    dt=new_dt,
                    t_0=new_t0,
                    diff_est=self.diff_est,
                )
            new_t = self.t[start:stop:step]
            return Trajectory(points=new_points, t=new_t, diff_est=self.diff_est)
        raise TypeError("Index must be an integer or a slice.")

    def __iter__(self) -> Iterator[TrajectoryPoint]:
        for i in range(len(self)):
            yield cast(TrajectoryPoint, self[i])

    @property
    def bounds(self) -> List[Tuple[float, float]]:
        """List[Tuple[float]] : List of tuples indicanting the min and
        max values of each dimension"""
        _bounds = []
        for dim in range(self.dim):
            min_bound = min(self.r.component(dim))
            max_bound = max(self.r.component(dim))
            _bounds.append((min_bound, max_bound))
        return _bounds

    @property
    def dim(self) -> int:
        """int : Trajectory spacial dimensions."""
        return self.r.shape[1]

    @property
    def delta_r(self) -> Vector:
        """Vector: Difference between each couple of consecutive points
        in the Trajectory."""
        return self.r.delta

    @property
    def delta_v(self) -> Vector:
        """Vector: Difference between each couple of consecutive sample
        in the velocity vector of the Trajectory."""
        return self.v.delta

    def recalculate_velocity(self) -> Vector:
        """
        Recalculates the velocity according time data or `dt` if time
        data is not available.

        Returns
        -------
        Vector
            Velocity vector.
        """
        self.__v = diff.estimate_velocity(self, **self.diff_est)
        return self.__v

    def recalculate_acceleration(self) -> Vector:
        """
        Recalculates the acceleration according time data or `dt` if time
        data is not available.

        Returns
        -------
        Vector
            Velocity vector.
        """
        self.__a = diff.estimate_accelereation(self, **self.diff_est)
        return self.__a

    @property
    def v(self) -> Vector:
        """Vector : Velocity vector"""
        if self.lazy and self.__v is not None:
            return self.__v
        return self.recalculate_velocity()

    @property
    def a(self) -> Vector:
        """Vector : Velocity vector"""
        if self.lazy and self.__a is not None:
            return self.__a
        return self.recalculate_acceleration()

    @property
    def t(self) -> Vector:
        """Vector : Time vector"""
        if self.__t is None:
            self.__t = Vector([self.t_0 + self.dt * i for i in range(len(self))])
        return self.__t

    def add_polar_offset(self, radius: float, angle: float) -> None:
        """
        Adds an offset given a point in polar coordinates.

        Parameters
        ----------
        radius : float
            Point's radius.
        angle : float
            Point's angle.

        Raises
        ------
        TypeError
            If the trajectory is not 2 dimensional.
        """
        if self.dim != 2:
            raise TypeError(
                "Polar offsets can only be applied on 2 dimensional trajectories"
            )

        # From cartesian to polar
        x, y = self.r.x, self.r.y
        rad, ang = np.hypot(x, y), np.arctan2(y, x)

        rad += radius
        ang += angle

        # From polar to cartesian
        x, y = rad * np.cos(ang), rad * np.sin(ang)
        self.r = Vector([x, y]).T

    def rotate_2d(self, angle: float):
        """
        Rotates the trajectory around the center coordinates [0,0]

        Parameters
        ----------
        angle : float
            Angle in radians to rotate the trajectory.
        """
        self.add_polar_offset(0, angle)

    def rotate2d(self, angle: float):
        """
        .. deprecated:: 0.10.0
            :func:`rotate2d` will be removed in a future version, use
            :func:`rotate_2d` instead.
        """
        warnings.warn(
            "rotate2d is deprecated and will be removed in a future version, "
            "use rotate_2d instead",
            DeprecationWarning,
        )
        self.rotate_2d(angle)

    def rotate_3d(self, angle: float, vector: Collection[float]):
        """
        Rotates the trajectory around a given vector.

        Parameters
        ----------
        vector : Collection[float]
            Vector to rotate the trajectory around.
        angle : float
            Angle in radians to rotate the trajectory.

        Raises
        ------
        TypeError
            If the trajectory is not 3 dimensional.
        ValueError
            If the vector has shape different than (3,).
        """
        if self.dim != 3:
            raise TypeError(
                "3D rotations can only be applied on 3 dimensional trajectories"
            )

        vec = Vector(vector)
        if vec.shape != (3,):
            raise ValueError("The vector must have shape (3,)")

        vec = vec / vec.norm
        v_x, v_y, v_z = vec[0], vec[1], vec[2]
        a_cos, a_sin = np.cos(angle), np.sin(angle)

        rot_matrix = np.array(
            [
                [
                    v_x * v_x * (1 - a_cos) + a_cos,
                    v_x * v_y * (1 - a_cos) - v_z * a_sin,
                    v_x * v_z * (1 - a_cos) + v_y * a_sin,
                ],
                [
                    v_x * v_y * (1 - a_cos) + v_z * a_sin,
                    v_y * v_y * (1 - a_cos) + a_cos,
                    v_y * v_z * (1 - a_cos) - v_x * a_sin,
                ],
                [
                    v_x * v_z * (1 - a_cos) - v_y * a_sin,
                    v_y * v_z * (1 - a_cos) + v_x * a_sin,
                    v_z * v_z * (1 - a_cos) + a_cos,
                ],
            ]
        )
        self.r = Vector(np.dot(self.r, rot_matrix))

    def rotate3d(self, angle: float, vector: Union[list, np.ndarray]):
        """
        .. deprecated:: 0.10.0
            :func:`rotate3d` will be removed in a future version, use
            :func:`rotate_3d` instead.
        """
        warnings.warn(
            "rotate3d is deprecated and will be removed in a future version, "
            "use rotate_3d instead",
            DeprecationWarning,
        )
        self.rotate_3d(angle, vector)

    def copy(self) -> Trajectory:
        """
        Returns a copy of the trajectory.

        Returns
        -------
        Trajectory
            Copy of the trajectory.
        """
        return Trajectory(
            points=self.r,
            t=self.__t,
            dt=self.__dt,
            lazy=self.lazy,
            diff_est=self.diff_est,
        )

    def _operable_with(
        self, other: Trajectory, threshold: Optional[float] = None
    ) -> bool:
        if self.r.shape != other.r.shape:
            return False

        threshold = _THRESHOLD if threshold is None else threshold
        self_time = self.t
        other_time = other.t
        diff = np.abs(np.subtract(self_time, other_time))
        return all(diff < threshold)

    def __iadd__(self, other):
        if isinstance(other, (int, float)):
            self.r += other
            return self

        if isinstance(other, (list, tuple, np.ndarray)):
            offset = np.array(other, dtype=float)
            if len(offset) != self.dim:
                raise ValueError(
                    "Offset must be the same shape as the other trajectory points"
                )
            self.r += offset
            return self

        if isinstance(other, Trajectory):
            if not self._operable_with(other):
                raise ValueError("Incompatible trajectories")
            self.r += other.r
            return self

        raise TypeError(
            "unsoported operation (+) between 'Trajectory' and "
            f"'{type(other).__name__}'"
        )

    def __isub__(self, other):
        if isinstance(other, (int, float)):
            self.r -= other
            return self

        if isinstance(other, (list, tuple, np.ndarray)):
            offset = np.array(other, dtype=float)
            if len(offset) != self.dim:
                raise ValueError(
                    "Offset must be the same shape as the other trajectory points"
                )
            self.r -= offset
            return self

        if isinstance(other, Trajectory):
            if not self._operable_with(other):
                raise ValueError("Incompatible trajectories")
            self.r -= other.r
            return self

        raise TypeError(
            "unsoported operation (-) between 'Trajectory' and "
            f"'{type(other).__name__}'"
        )

    def __add__(self, other):
        traj = self.copy()
        traj += other
        return traj

    def __sub__(self, other):
        traj = self.copy()
        traj -= other
        return traj

    def __radd__(self, other):
        return self + other

    def __rsub__(self, other):
        return self - other

    def __imul__(self, other):
        if isinstance(other, (int, float)):
            self.r *= other
            return self
        raise TypeError(
            "unsoported operation (*) between 'Trajectory' and "
            f"'{type(other).__name__}'"
        )

    def __mul__(self, other):
        traj = self.copy()
        traj *= other
        return traj

    def __rmul__(self, other):
        return self * other

    def turning_angles(
        self, accumulate=False, degrees=False, centered=False, wrap=True
    ):
        """
        Return the sequence of turning angles that forms the trajectory.

        Parameters
        ----------
        traj : Trajectory
            Input trajectory.
        accumulate : bool, optional
            If True, turning angles are measured with respect to an axis
            defined by the initial velocity (i.e., angles between initial
            and current velocity). Otherwise, relative turning angles
            are computed (i.e., angles between succesive velocity
            vectors). By default False.
        degrees : bool, optional
            If True, angles are given in degrees. Otherwise, the units
            are radians. By default False.
        centered : bool, optional
            If True, angles are wrapped on the interval ``[-pi, pi]``.
            Otherwise, the interval ``[0, 2*pi]`` is chosen. By default
            False.
        wrap : bool, optional
            If True, angles are wrapped in a certain interval (depending
            on ``centered`` param). By default True.

        Returns
        -------
        np.ndarray
            Turning angles where each position in the array correspond
            to a given time instant.
        """
        d_r = self.delta_r
        d_x, d_y = d_r.x, d_r.y
        theta = np.arctan2(d_y, d_x)

        if not accumulate:
            theta = np.ediff1d(theta)  # Relative turning angles
        else:
            theta -= theta[0]  # Accumulative turning angles

        if degrees:
            theta = np.rad2deg(theta)

        if not wrap:
            return theta

        discont = 360 if degrees else 2 * np.pi
        if not centered:
            return theta % discont

        discont_half = discont / 2
        return -((discont_half - theta) % discont - discont_half)

    def _save_json(self, path: Union[str, Path]) -> None:
        def convert_to_list(vec: Optional[Vector]):
            if vec is None:
                return vec
            if len(vec.shape) == 1:
                return list(vec)
            return {d: list(v) for d, v in enumerate(vec)}

        diff_est = {
            "method": self.diff_est.get("method", diff.DiffMethod.LINEAR_DIFF).value,
            "window_type": self.diff_est.get("window", diff.WindowType.CENTRAL).value,
            "accuracy": self.diff_est.get("accuracy", 1),
        }

        json_dict = {
            "id": self.traj_id,
            "dt": self.__dt,
            "r": convert_to_list(self.r.T),
            "t": convert_to_list(self.__t),
            "diff_est": diff_est,
        }
        with open(str(path), "w", encoding="utf-8") as traj_file:
            json.dump(json_dict, traj_file)

    def _save_csv(self, path: Union[str, Path]) -> None:
        with open(str(path), "w", newline="", encoding="utf-8") as traj_file:
            writer = csv.writer(traj_file, delimiter=",")
            writer.writerow([self.traj_id, self.__dt, self.dim])

            default_diff_method = diff.DiffMethod.LINEAR_DIFF
            default_diff_window = diff.WindowType.CENTRAL
            default_diff_accuracy = 1
            method = self.diff_est.get("method", default_diff_method).value
            window = self.diff_est.get("window", default_diff_window).value
            accuracy = self.diff_est.get("accuracy", default_diff_accuracy)
            writer.writerow([method, window, accuracy])

            for t_p in self:
                row = np.hstack([t_p.r, t_p.t])
                writer.writerow(row)

    def save(
        self,
        file_name: str,
        path: str = ".",
        file_type: str = "json",
        overwrite: bool = True,
    ):
        """
        .. deprecated:: 0.10.0
            :func:`save` will be removed in a future version, use a Serializer
            from ``yupi.core`` instead (e.g., JSONSerializer).

        Saves the trajectory to disk.

        Parameters
        ----------
        file_name : str
            Name of the file.
        path : str
            Path where to save the trajectory, by default ``'.'``.
        file_time : str
            Type of the file, by default ``json``.

            The only types avaliable are: ``json`` and ``csv``.
        overwrite : bool
            Wheter or not to overwrite the file if it already exists,
            by default True.

        Raises
        ------
        ValueError
            If ``override`` parameter is ``False`` and the file already
            exists.
        ValueError
            If ``file_type`` is not ``json`` or ``csv``.

        Examples
        --------
        >>> t = Trajectory(x=[0.37, 1.24, 1.5])
        >>> t.save('my_track')
        """
        warnings.warn(
            "`save` is deprecated and will be removed in a future version, use a "
            "Serializer from `yupi.core` instead (e.g., JSONSerializer).",
            DeprecationWarning,
        )

        # Build full path
        full_path = Path(path) / Path(f"{file_name}.{file_type}")

        # Check file existance
        if not overwrite and full_path.exists():
            raise FileExistsError(f"File '{str(full_path)}' already exist")

        if file_type == "json":
            self._save_json(full_path)
        elif file_type == "csv":
            self._save_csv(full_path)
        else:
            raise ValueError(f"Invalid export file type '{file_type}'")

    @staticmethod
    def save_trajectories(
        trajs: List[Trajectory],
        folder_path: str = ".",
        file_type: str = "json",
        overwrite: bool = True,
    ):
        """
        Saves a list of trajectories to disk. Each Trajectory object
        will be saved in a separate file inside the given folder.

        Parameters
        ----------
        trajs : list[Trajectory]
            List of Trajectory objects that will be saved.
        folder_path : str
            Path where to save all the trajectory, by default ``'.'``.
        file_type : str
            Type of the file, by default ``jon``.

            The only types avaliable are: ``json`` and ``csv``.
        overwrite : bool
            Wheter or not to overwrite the file if it already exists,
            by default True.

        Examples
        --------
        >>> t1 = Trajectory(x=[0.37, 1.24, 1.5])
        >>> t2 = Trajectory(x=[1, 2, 3], y=[3, 4, 5])
        >>> Trajectory.save_trajectories([t1, t2])
        """
        for i, traj in enumerate(trajs):
            path = str(Path(folder_path))
            name = str(Path(f"trajectory_{i}"))
            traj.save(name, path, file_type, overwrite)

    @staticmethod
    def _load_json(path: str):
        with open(path, "r", encoding="utf-8") as traj_file:
            data = json.load(traj_file)

            traj_id = data["id"]
            t, dt = data["t"], data["dt"]
            axes = list(data["r"].values())
            diff_est = data.get("diff_est", None)
            if diff_est is None:
                diff_est = Trajectory.general_diff_est
            else:
                diff_est["method"] = diff.DiffMethod(diff_est["method"])
                diff_est["window_type"] = diff.WindowType(diff_est["window_type"])

            return Trajectory(axes=axes, t=t, dt=dt, traj_id=traj_id, diff_est=diff_est)

    @staticmethod
    def _load_csv(path: str):
        with open(path, "r", encoding="utf-8") as traj_file:

            def check_empty_val(val, cast_value=True) -> Union[None, float]:
                if val == "":
                    return None
                return float(val) if cast_value else val

            r: List[List[float]] = []
            t: List[float] = []
            traj_id: Optional[str] = None
            dt, dim = 1.0, 1
            diff_est = Trajectory.general_diff_est

            for i, row in enumerate(csv.reader(traj_file)):
                if i == 0:
                    traj_id = row[0] if row[0] != "" else None
                    dt = check_empty_val(row[1])
                    dim = int(row[2])
                    r = [[] for _ in range(dim)]
                    continue

                if i == 1:
                    diff_est = {
                        "method": diff.DiffMethod(int(row[0])),
                        "window_type": diff.WindowType(int(row[1])),
                        "accuracy": int(row[2]),
                    }
                    continue

                for j in range(dim):
                    r[j].append(float(row[j]))

                t.append(float(row[-1]))

            return Trajectory(axes=r, t=t, dt=dt, traj_id=traj_id, diff_est=diff_est)

    @staticmethod
    def load(file_path: str):
        """
        .. deprecated:: 0.10.0
            :func:`load` will be removed in a future version, use a Serializer
            from ``yupi.core`` instead (e.g., JSONSerializer).

        Loads a trajectory

        Parameters
        ----------
        file_path : str
            Path of the trajectory file

        Returns
        -------
        Trajectory
            Loaded Trajectory object.

        Raises
        ------
        ValueError
            If ``file_path`` is a non existing path.
        ValueError
            If ``file_path`` is a not a file.
        ValueError
            If ``file_path`` extension is not ``json`` or ```csv``.
        """

        warnings.warn(
            "`load` is deprecated and will be removed in a future version, use a "
            "Serializer from `yupi.core` instead (e.g., JSONSerializer).",
            DeprecationWarning,
        )

        path = Path(file_path)
        # Check valid path
        if not path.exists():
            raise ValueError("Path does not exist.")
        if not path.is_file():
            raise ValueError("Path must be a file.")

        file_type = path.suffix

        try:
            if file_type == ".json":
                return Trajectory._load_json(file_path)
            if file_type == ".csv":
                return Trajectory._load_csv(file_path)
            raise ValueError("Invalid file type.")
        except (json.JSONDecodeError, KeyError, ValueError, IndexError) as exc:
            raise LoadTrajectoryError(str(path)) from exc

    @staticmethod
    def load_folder(folder_path=".", recursively: bool = False):
        """
        Loads all the trajectories from a folder.

        Parameters
        ----------
        folder_path : str
            Path of the trajectories folder.
        recursively : bool
            If True then subfolders are analized recursively, by
            default False.

        Returns
        -------
        List[Trajectory]
            List of the loaded trajectories.
        """
        trajs = []
        for root, _, files in os.walk(folder_path):
            for file in files:
                path = str(Path(root) / Path(file))
                try:
                    trajs.append(Trajectory.load(path))
                except LoadTrajectoryError as load_exception:
                    logging.warning("Ignoring: '%s'", load_exception.path)
            if not recursively:
                break
        return trajs



from yupi.core.serializers.csv_serializer import CSVSerializer
from yupi.core.serializers.json_serializer import JSONSerializer
from yupi.core.serializers.serializer import Serializer

__all__ = [
    "CSVSerializer",
    "JSONSerializer",
    "Serializer",
]



"""
This contains a series of decorators that check the consistency of
the trajectories according to different criteria.
"""

from typing import List

import numpy as np

from yupi.trajectory import _THRESHOLD, Trajectory


def check_uniform_time_spaced(func):
    """Check that the trajectories are uniformly time-spaced."""

    def wrapper(trajs: List[Trajectory], *args, **kwargs):
        if any((abs(t.dt_std - 0) > _THRESHOLD for t in trajs)):
            raise ValueError("All trajectories must be uniformly time spaced")
        return func(trajs, *args, **kwargs)

    wrapper.__doc__ = func.__doc__
    return wrapper


def check_same_dt(func):
    """Check that the trajectories have the same dt."""

    def wrapper(trajs: List[Trajectory], *args, **kwargs):
        if trajs:
            dt = trajs[0].dt
            if any((abs(t.dt - dt) > _THRESHOLD for t in trajs)):
                raise ValueError("All trajectories must have the same 'dt'")
        return func(trajs, *args, **kwargs)

    wrapper.__doc__ = func.__doc__
    return wrapper


def check_same_dim(func):
    """Check that the trajectories have the same dimension."""

    def wrapper(trajs: List[Trajectory], *args, **kwargs):
        if trajs:
            dim = trajs[0].dim
            if any((t.dim != dim for t in trajs)):
                raise ValueError("All trajectories must have the same dimensions")
        return func(trajs, *args, **kwargs)

    wrapper.__doc__ = func.__doc__
    return wrapper


def check_exact_dim(dim):
    """
    Check that the trajectories have the same given dimension.

    Parameters
    ----------
    dim : int
        The dimension to check.
    """

    def _check_exact_dim_decorator(func):
        def wrapper(trajs: List[Trajectory], *args, dim=dim, **kwargs):
            if any((t.dim != dim for t in trajs)):
                raise ValueError(f"All trajectories must be {dim}-dimensional")
            return func(trajs, *args, **kwargs)

        wrapper.__doc__ = func.__doc__
        return wrapper

    return _check_exact_dim_decorator


def check_same_r0(func):
    """Check that the trajectories have the same initial position."""

    def wrapper(trajs: List[Trajectory], *args, **kwargs):
        if trajs:
            r_0 = trajs[0].r[0]
            if any((abs(t.r[0] - r_0) > _THRESHOLD for t in trajs)):
                raise ValueError("All trajectories must have the same initial position")
        return func(trajs, *args, **kwargs)

    wrapper.__doc__ = func.__doc__
    return wrapper


def check_same_lenght(func):
    """Check that the trajectories have the same lenght."""

    def wrapper(trajs: List[Trajectory], *args, **kwargs):
        if trajs:
            length = len(trajs[0])
            if any((abs(len(t) - length) > _THRESHOLD for t in trajs)):
                raise ValueError("All trajectories must have the same length")
        return func(trajs, *args, **kwargs)

    wrapper.__doc__ = func.__doc__
    return wrapper


def check_same_t(func):
    """Check that the trajectories have the same time data."""

    def wrapper(trajs: List[Trajectory], *args, **kwargs):
        if trajs:
            zipped_trajs = zip(trajs[:-1], trajs[1:])
            if not all(
                (np.allclose(t0.t, t1.t, atol=_THRESHOLD) for t0, t1 in zipped_trajs)
            ):
                raise ValueError("All trajectories must have the same 't' values")
        return func(trajs, *args, **kwargs)

    wrapper.__doc__ = func.__doc__
    return wrapper



"""
This constains resampling functions for trajectories.
"""
from typing import Collection, List, Optional, Union

import numpy as np

from yupi import Trajectory
from yupi._differentiation import _get_coeff


def _get_k_value_neighbors(val: float, data, k: int, _from: int):
    lower_bound, upper_bound = _from, _from
    for _ in range(k):
        look_forwards = (
            data[upper_bound] - val <= val - data[lower_bound] or lower_bound == 0
        )
        if look_forwards and upper_bound < len(data) - 1:
            upper_bound = min(upper_bound + 1, len(data) - 1)
        else:
            lower_bound = max(lower_bound - 1, 0)
    return lower_bound, upper_bound


def _interpolate_axis(axis_data, old_t, new_t, order):
    new_t_idxs = np.searchsorted(old_t, new_t)
    assert isinstance(new_t_idxs, np.ndarray)
    new_dim = np.empty(len(new_t))
    for i, new_t_idx in enumerate(new_t_idxs):
        val = new_t[i]
        min_neighbor, max_neighbor = _get_k_value_neighbors(
            val, old_t, order, new_t_idx
        )
        alphas = old_t[min_neighbor : max_neighbor + 1]
        _coeff = _get_coeff(val, alphas, M=1)
        new_dim[i] = np.sum(
            _coeff[0, len(alphas) - 1, :] * axis_data[min_neighbor : max_neighbor + 1]
        )
    return new_dim


def resample(
    traj: Trajectory,
    new_dt: Optional[float] = None,
    new_t: Optional[Collection[float]] = None,
    new_traj_id: Optional[str] = None,
    order: int = 1,
):
    """
    Resamples a trajectory to a new dt or a new array of time.

    One of ``new_dt`` or ``new_t`` must be specified.

    Parameters
    ----------
    traj : Trajectory
        Input trajectory.
    new_dt: Optional[float]
        New dt. By default None.
    new_t: Optional[Collection[float]]
        New sample rate or array of time. By default None.
    new_traj_id : Optional[str]
        New trajectory ID. By default None.
    order : int, optional
        How many points to use for the interpolation of each value. By default 2.

    Returns
    -------
    Trajectory
        Output trajectory.

    Raises
    ------
    ValueError
        If neither ``new_dt`` nor ``new_t`` is specified.
    ValueError
        If both ``new_dt`` and ``new_t`` are specified.
    """

    if new_t is not None and new_dt is not None:
        raise ValueError("new_t and new_dt cannot be both specified")
    if new_t is None and new_dt is None:
        raise ValueError("new_t or new_dt must be specified")

    from_dt = new_dt is not None

    new_t = (
        traj.t[0] + np.arange(0, traj.t[-1], new_dt)
        if new_dt is not None
        else np.array(new_t)
    )
    new_dims: List[Collection[float]] = []
    old_t = traj.t

    for dim in range(traj.dim):
        dim_data = traj.r.component(dim)
        new_dim = _interpolate_axis(dim_data, old_t, new_t, order)
        new_dims.append(new_dim)

    if from_dt:
        return Trajectory(
            axes=new_dims,
            dt=new_dt,
            traj_id=new_traj_id,
            diff_est=traj.diff_est,
        )
    return Trajectory(
        axes=new_dims,
        t=new_t,
        traj_id=new_traj_id,
        diff_est=traj.diff_est,
    )


def subsample(traj: Trajectory, step: int = 1, new_traj_id: Optional[str] = None):
    """
    Sample the trajectory ``traj`` by removing evenly spaced
    points according to ``step``.

    Parameters
    ----------
    traj : Trajectory
        Input trajectory.
    step : int, optional
        Number of sample points or period. By default 1.
    new_traj_id : Optional[str]
        New trajectory ID. By default None.

    Returns
    -------
    Trajectory
        Output trajectory.
    """

    points = traj.r[::step]
    t = traj.t[::step] if traj.t is not None else None

    subsampled_traj = Trajectory(
        points=points,
        t=t,
        dt=step * traj.dt,
        traj_id=new_traj_id,
        diff_est=traj.diff_est,
    )
    return subsampled_traj



"""
This contains styling utilities for the library plots.
"""

from typing import Optional

import matplotlib.pyplot as plt

LINE = "-"
DOTTED = "o"
LINE_DOTTED = "-o"

# Main colors
BLUE = "#3489b1"
MAGENTA = "#a14383"
YELLOW = "#e5d935"
RED = "#af3c3c"
GREEN = "#54ac43"
ORANGE = "#e88d26"
MID_BLUE = "#3c4baf"

#  Light colors
LIGHT_BLUE = "#99d2ec"
LIGHT_MAGENTA = "#eaa0d2"
LIGHT_YELLOW = "#fdf584"
LIGHT_RED = "#ea8080"
LIGHT_GREEN = "#a6ec98"
LIGHT_ORANGE = "#f7c790"

YUPI_COLORS = [BLUE, MAGENTA, YELLOW, RED, GREEN, ORANGE, MID_BLUE]

YUPI_LIGHT_COLORS = [
    LIGHT_BLUE,
    LIGHT_MAGENTA,
    LIGHT_YELLOW,
    LIGHT_RED,
    LIGHT_GREEN,
    LIGHT_ORANGE,
]


def _plot_basic_properties(func):
    def wrapper(
        *args,
        title: str = "",
        xlabel: str = "",
        ylabel: str = "",
        grid: bool = False,
        show: bool = True,
        legend: bool = False,
        xscale: Optional[str] = None,
        yscale: Optional[str] = None,
        xlim: Optional[tuple] = None,
        ylim: Optional[tuple] = None,
        **kwargs,
    ):
        func(*args, **kwargs)
        plt.grid(grid)
        plt.title(title)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        if legend:
            plt.legend()
        if xlim is not None:
            plt.xlim(xlim)
        if ylim is not None:
            plt.ylim(ylim)
        if xscale is not None:
            plt.xscale(xscale)
        if yscale is not None:
            plt.yscale(yscale)
        if show:
            plt.show()

    wrapper.__doc__ = func.__doc__
    return wrapper



"""
This contains the base class for all serializers.
"""

from __future__ import annotations

import abc
from pathlib import Path
from typing import Optional

import yupi


class Serializer(abc.ABC):
    """
    Abstract class for trajectory files.
    """

    @staticmethod
    @abc.abstractmethod
    def save(
        traj: yupi.Trajectory, file_name: str, overwrite: bool = False, **kwargs
    ) -> None:
        """
        Saves a trajectory to a file.

        Parameters
        ----------
        traj : Trajectory
            The trajectory to be saved.
        file_name : str
            The name of the file to save.
        overwrite : bool
            If True, overwrites the file if it already exists.
        kwargs
            Additional keyword arguments.
        """

    @staticmethod
    @abc.abstractmethod
    def load(file_name: str, **kwargs) -> yupi.Trajectory:
        """
        Loads a trajectory from a file.

        Parameters
        ----------
        file_name : str
            The name of the file to loaded.
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        Trajectory
            The trajectory loaded from the file.
        """

    @staticmethod
    def check_save_path(
        file_name: str, overwrite: bool, extension: Optional[str]
    ) -> None:
        """
        Checks if the file can be saved.

        Parameters
        ----------
        file_name : str
            The name of the file to save.
        overwrite : bool
            If True, overwrites the file if it already exists.
        extension : Optional[str]
            If given, it checks that the file has the given extension.
        """
        _path = Path(file_name)
        if extension is not None and _path.suffix != extension:
            raise ValueError(
                f"File extension must be '{extension}', not {_path.suffix}"
            )

        if _path.exists() and not overwrite:
            raise FileExistsError(f"File '{file_name}' already exists.")

        _path.parent.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def check_load_path(file_name: str, extension: Optional[str]) -> None:
        """
        Checks if the file can be loaded.

        Parameters
        ----------
        file_name : str
            The name of the file to loaded.
        extension : Optional[str]
            If given, it checks that the file has the given extension.
        """
        _path = Path(file_name)
        if extension is not None and _path.suffix != extension:
            raise ValueError(
                f"File extension must be '{extension}', not {_path.suffix}"
            )

        if not _path.exists():
            raise FileNotFoundError(f"File '{file_name}' not found.")



"""
This contains all the affine estimator related functions.
"""

import logging
from typing import Optional, Tuple

import cv2
import nudged
import numpy as np

AffineParams = Tuple[float, float, float, float]
"""Affine params: theta, t_x, t_y, scale."""

# ShiTomasi corner detection
FEATURE_PARAMS = dict(maxCorners=30, qualityLevel=0.6, minDistance=30, blockSize=100)

# Lucas Kanade optical flow
LK_PARAMS = dict(
    winSize=(20, 20),
    maxLevel=15,
    criteria=(cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 200, 0.05),
)


def _rot_matrix(theta: float, inverse: bool = False) -> np.ndarray:
    rot_mat = np.array(
        [[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]]
    )
    return rot_mat.T if inverse else rot_mat


def _affine_matrix(
    theta: float, t_x: float, t_y: float, scale: float = 1.0, inverse: bool = False
) -> np.ndarray:
    rot_mat = _rot_matrix(theta, inverse)
    rot_mat = (scale * np.identity(2)) @ rot_mat
    shift = np.array([t_x, t_y])
    aff_mat = np.hstack([rot_mat, shift[None, :].T])
    return aff_mat


def _estimate_params(
    p_1: np.ndarray, p_2: np.ndarray
) -> Tuple[float, float, float, float]:
    transf = nudged.estimate(p_1, p_2)
    t_x, t_y = transf.get_translation()
    theta, scale = transf.get_rotation(), transf.get_scale()
    return theta, t_x, t_y, scale


def _get_p3(
    p_1: np.ndarray, theta: float, t_x: float, t_y: float, scale: float = 1.0
) -> np.ndarray:
    aff_mat = _affine_matrix(theta, t_x, t_y, scale)
    p_3 = np.array([aff_mat @ [x1, y1, 1] for (x1, y1) in p_1])
    return p_3


def _get_r(p_1: np.ndarray, p_2: np.ndarray) -> np.ndarray:
    x, y = (p_1 - p_2).T
    return np.sqrt(x**2 + y**2)


def _get_mask_r(r: np.ndarray, quantile: float = 1.5) -> np.ndarray:
    # Median and standard deviation of distance population
    r_median, r_std = np.median(r), np.std(r)
    # Cutoff used to filter far features points
    bound = max(1, quantile * r_std)
    # Get mask for nearby points
    mask = r < r_median + bound
    return mask


def _delete_far_points(
    p_1: np.ndarray,
    p_2: np.ndarray,
    p_3: Optional[np.ndarray] = None,
    quantile: float = 1.5,
) -> Tuple[np.ndarray, np.ndarray]:
    # Distance amoung p1 and p2, or p2 and p3
    r = _get_r(p_1, p_2) if p_3 is None else _get_r(p_2, p_3)
    # Mask that filters outliers for a given quantile value
    mask = _get_mask_r(r, quantile)
    # Delete far points
    p_1, p_2 = p_1[mask], p_2[mask]
    return p_1, p_2


def _estimate_matrix(
    p_1: np.ndarray,
    p_2: np.ndarray,
    quantile1: Optional[float] = None,
    quantile2: Optional[float] = None,
) -> Tuple[Tuple[np.ndarray, np.ndarray, np.ndarray], AffineParams]:
    # Validate tracked features deletting outliers
    if quantile1 is not None:
        p_1, p_2 = _delete_far_points(p_1, p_2, quantile=quantile1)

    # Estimate matrix parameters and transformed features
    affine_params = _estimate_params(p_1, p_2)
    p_3 = _get_p3(p_1, *affine_params)

    # Delete outliers considering transformed features
    if quantile2 is not None:
        p_1, p_2 = _delete_far_points(p_1, p_2, p_3, quantile=quantile2)
        affine_params = _estimate_params(p_1, p_2)
        p_3 = _get_p3(p_1, *affine_params)

    points = p_1, p_2, p_3
    return points, affine_params


def _get_rmse(p_2: np.ndarray, p_3: np.ndarray) -> float:
    x, y = (p_2 - p_3).T
    return np.sqrt(np.mean(x**2 + y**2))


def _get_affine(
    img1: np.ndarray,
    img2: np.ndarray,
    region: Tuple[int, int, int, int],
    mask: Optional[np.ndarray] = None,
) -> Tuple[Tuple[np.ndarray, np.ndarray, np.ndarray], AffineParams, Optional[float]]:
    x_0, x_f, y_0, y_f = region

    # Get main regions
    img1_region = img1[y_0:y_f, x_0:x_f, :]
    img2_region = img2[y_0:y_f, x_0:x_f, :]

    if mask is not None:
        mask = mask[y_0:y_f, x_0:x_f]

    # Convert to grayscale
    img1_region_gray = cv2.cvtColor(img1_region, cv2.COLOR_BGR2GRAY)
    img2_region_gray = cv2.cvtColor(img2_region, cv2.COLOR_BGR2GRAY)

    # Equalize histograms to improve contrast
    img1_region_gray = cv2.equalizeHist(img1_region_gray)
    img2_region_gray = cv2.equalizeHist(img2_region_gray)

    # Track good features
    p1_region = cv2.goodFeaturesToTrack(img1_region_gray, mask=mask, **FEATURE_PARAMS)
    p2_region, st, _ = cv2.calcOpticalFlowPyrLK(  # pylint: disable=invalid-name
        img1_region_gray, img2_region_gray, p1_region, None, **LK_PARAMS
    )

    # Change origin and select tracked points
    p1_frame, p2_frame = p1_region + [x_0, y_0], p2_region + [x_0, y_0]
    p1_good, p2_good = p1_frame[st == 1], p2_frame[st == 1]

    # Cancel estimation if no good points were found or tracked
    if p1_good.size == 0:
        logging.error("No good points were found or sucessfully tracked.")
        return 3 * (0,), 3 * (0,), None

    # Estimate points and matrix
    p_good, affine_params = _estimate_matrix(
        p1_good, p2_good, quantile1=1.5, quantile2=1.5
    )
    p1_good, p2_good, p3_good = p_good

    # Mean square error
    err = _get_rmse(p2_good, p3_good)

    return p_good, affine_params, err



"""
This module contains different statistical models to
generate trajectories given certain statistical constrains.

All the resources of this module should be imported directly
from ``yupi.generators``.
"""

from yupi.generators._generators import (
    DiffDiffGenerator,
    Generator,
    LangevinGenerator,
    RandomWalkGenerator,
)

__all__ = ["Generator", "RandomWalkGenerator", "LangevinGenerator", "DiffDiffGenerator"]



"""
All the resources of the library should be imported directly
from one of the aforementioned modules.
"""

import logging
import warnings

from yupi._differentiation import DiffMethod, WindowType
from yupi.features import Features
from yupi.trajectory import Trajectory, TrajectoryPoint
from yupi.vector import Vector

warnings.filterwarnings("default", category=DeprecationWarning, module="yupi")
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(message)s",
    datefmt="%H:%M:%S",
)

__all__ = [
    "Trajectory",
    "TrajectoryPoint",
    "Features",
    "Vector",
    "DiffMethod",
    "WindowType",
]

__version__ = "0.11.2"



"""
JSON trajctory serializer.
"""

import json
import logging
from typing import List

import yupi._differentiation as diff
from yupi import Trajectory
from yupi.core.serializers.serializer import Serializer
from yupi.exceptions import LoadTrajectoryError


class JSONSerializer(Serializer):
    """
    Handles trajectory files in JSON format.
    """

    @staticmethod
    def save(
        traj: Trajectory, file_name: str, overwrite: bool = False, **kwargs
    ) -> None:
        """
        Writes a trajectory to a file.

        Parameters
        ----------
        traj : Trajectory
            The trajectory to write to the file.
        file_name : str
            The name of the file to write.
        overwrite : bool
            If True, overwrites the file if it already exists.
        kwargs
            Additional arguments to pass to the ``open`` function.

            Encoding is set to UTF-8 as default.
        """
        JSONSerializer.check_save_path(
            file_name, overwrite=overwrite, extension=".json"
        )

        json_dict = JSONSerializer.to_json(traj)
        encoding = "utf-8" if "encoding" not in kwargs else kwargs.pop("encoding")
        with open(file_name, "w", encoding=encoding, **kwargs) as traj_file:
            json.dump(json_dict, traj_file)

    @staticmethod
    def save_ensemble(
        trajs: List[Trajectory], file_name: str, overwrite: bool = False, **kwargs
    ) -> None:
        """
        Writes an ensemble to a file.

        The main difference with the ``save`` method is that all the
        trajectories are saved in a single file.

        Parameters
        ----------
        trajs : List[Trajectory]
            The ensemble to write to the file.
        file_name : str
            The name of the file to write.
        overwrite : bool
            If True, overwrites the file if it already exists.
        kwargs
            Additional arguments to pass to the ``open`` function.

            Encoding is set to UTF-8 as default.
        """
        JSONSerializer.check_save_path(
            file_name, overwrite=overwrite, extension=".json"
        )

        json_dicts = [JSONSerializer.to_json(traj) for traj in trajs]
        encoding = "utf-8" if "encoding" not in kwargs else kwargs.pop("encoding")
        with open(file_name, "w", encoding=encoding, **kwargs) as traj_file:
            json.dump(json_dicts, traj_file)

    @staticmethod
    def load(file_name: str, **kwargs) -> Trajectory:
        """
        Loads a trajectory from a file.

        Parameters
        ----------
        file_name : str
            The name of the file to loaded.
        kwargs : dict
            Additional keyword arguments.

            Encoding is set to UTF-8 as default.

        Returns
        -------
        Trajectory
            The trajectory loaded from the file.
        """
        JSONSerializer.check_load_path(file_name, extension=".json")

        encoding = "utf-8" if "encoding" not in kwargs else kwargs.pop("encoding")
        with open(file_name, "r", encoding=encoding, **kwargs) as file:
            data = json.load(file)

            if "axes" not in data and "r" not in data:
                raise LoadTrajectoryError(file_name, "No position data found.")
            if "dt" not in data and "t" not in data:
                raise LoadTrajectoryError(file_name, "No time data found.")
            return JSONSerializer.from_json(data)

    @staticmethod
    def load_ensemble(file_name: str, **kwargs) -> List[Trajectory]:
        """
        Loads an ensemble from a file.

        The main difference with the ``load`` method is that all the
        trajectories are loaded from a single file.

        Parameters
        ----------
        file_name : str
            The name of the file to loaded.
        kwargs : dict
            Additional keyword arguments.

            Encoding is set to UTF-8 as default.

        Returns
        -------
        List[Trajectory]
            The ensemble loaded from the file.
        """
        JSONSerializer.check_load_path(file_name, extension=".json")

        encoding = "utf-8" if "encoding" not in kwargs else kwargs.pop("encoding")
        with open(file_name, "r", encoding=encoding, **kwargs) as file:
            data = json.load(file)

            if any("axes" not in traj and "r" not in traj for traj in data):
                raise LoadTrajectoryError(
                    file_name, "No position data found for one or more trajectories."
                )
            if any("dt" not in traj and "t" not in traj for traj in data):
                raise LoadTrajectoryError(
                    file_name, "No time data found for one or more trajectories."
                )
            return [JSONSerializer.from_json(traj) for traj in data]

    @staticmethod
    def to_json(traj: Trajectory) -> dict:
        """
        Converts a trajectory to a JSON dictionary.

        Parameters
        ----------
        traj : Trajectory
            The trajectory to convert.

        Returns
        -------
        dict
            The JSON dictionary.
        """

        method = Trajectory.general_diff_est.get("method", diff.DiffMethod.LINEAR_DIFF)
        window = Trajectory.general_diff_est.get("window_type", diff.WindowType.CENTRAL)
        accuracy = Trajectory.general_diff_est.get("accuracy", 1)
        diff_est = {
            "method": traj.diff_est.get("method", method).value,
            "window_type": traj.diff_est.get("window_type", window).value,
            "accuracy": traj.diff_est.get("accuracy", accuracy),
        }

        json_dict = {
            "axes": traj.r.T.tolist(),
            "id": traj.traj_id,
            "diff_est": diff_est,
        }
        if traj.dt_std == 0:
            json_dict["dt"] = traj.dt
            json_dict["t_0"] = traj.t_0
        else:
            json_dict["t"] = traj.t.tolist()
        return json_dict

    @staticmethod
    def from_json(json_traj: dict) -> Trajectory:
        """
        Converts a JSON dictionary to a trajectory.

        Parameters
        ----------
        json_traj : dict
            The JSON dictionary to convert.

        Returns
        -------
        Trajectory
            The trajectory.
        """
        axes = json_traj.get("axes", None)
        if axes is None:
            logging.warning(
                "Trajectory will be loaded but it seems to be saved in an old format. "
                "Please consider updating it by using the JSONSerializer.save method. "
                "Older format won't be supported in a future."
            )
            axes = list(json_traj["r"].values())
        traj_id = json_traj["id"] if json_traj["id"] is not None else ""

        diff_est = json_traj.get("diff_est", None)
        if diff_est is None:
            diff_est = Trajectory.general_diff_est
        else:
            diff_est["method"] = diff.DiffMethod(diff_est["method"])
            diff_est["window_type"] = diff.WindowType(diff_est["window_type"])

        t = json_traj.get("t", None)
        dt = json_traj.get("dt", None)
        t_0 = json_traj.get("t_0", 0.0)

        return Trajectory(
            axes=axes, t=t, dt=dt, t_0=t_0, traj_id=traj_id, diff_est=diff_est
        )



"""
This contains specific exceptions related to the library.
"""


class YupiException(Exception):
    """Generic exception for yupi package"""


class TrajectoryError(YupiException):
    """Generic exception for handling trajectory errors"""


class LoadTrajectoryError(TrajectoryError):
    """Error while loading a trajectory"""

    def __init__(self, path: str, reason: str = ""):
        self.path = path
        self.message = f"File '{self.path}' is not a valid trajectory"
        if reason:
            self.message += f": {reason}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



"""
This contains filtering functions for the trajectories.
"""

from typing import Optional

import numpy as np

from yupi.trajectory import Trajectory


def exp_convolutional_filter(
    traj: Trajectory, gamma: float, new_traj_id: Optional[str] = None
):
    """
    Returns a smoothed version of the trajectory `traj`
    by taking a weighted average over past values.

    Parameters
    ----------
    traj : Trajectory
        Input trajectory.
    gamma : float
        Inverse of the characteristic time window of
        the average.
    new_traj_id : Optional[str]
        New trajectory ID. By default None.

    Returns
    -------
    Trajectory
        Smoothed trajectory.
    """

    r = traj.r
    dt = np.ediff1d(traj.t)
    new_r = np.zeros_like(r)
    for i in range(len(traj) - 1):
        new_r[i + 1] = new_r[i] - gamma * (new_r[i] - r[i]) * dt[i]

    smooth_traj = Trajectory(
        points=new_r, t=traj.t, traj_id=new_traj_id, diff_est=traj.diff_est
    )
    return smooth_traj



"""
This contains the Vector structure used across the library to store data.
"""

from __future__ import annotations

import warnings
from typing import Union

import numpy as np
from numpy.linalg.linalg import norm as nrm


class Vector(np.ndarray):
    """Represents a vector"""

    def __new__(cls, arr, dtype=None, copy=False):
        vec = np.asarray(arr, dtype=dtype)
        if copy:
            vec = vec.copy()
        return vec.view(cls)

    @property
    def norm(self) -> Union[Vector, float]:
        """Vector : Calculates the norm of the vector. If the vector
        is alist of vectors then the norm of each item is calculated"""
        if len(self.shape) < 2:
            return float(nrm(self))
        return Vector(nrm(self, axis=1))

    @property
    def delta(self) -> Vector:
        """Vector : Calculates the differnece between each item"""
        return Vector(np.diff(self, axis=0))

    @property
    def x(self) -> Vector:
        """Vector : X component of all vector items"""
        return self.component(0)

    @property
    def y(self) -> Vector:
        """Vector : Y component of all vector items"""
        return self.component(1)

    @property
    def z(self) -> Vector:
        """Vector : Z component of all vector items"""
        return self.component(2)

    def component(self, dim) -> Vector:
        """
        Extract a given component from all vector items.

        Parameters
        ----------
        dim : int
            Component index.

        Returns
        -------
        Vector
            Component extracted.

        Raises
        ------
        TypeError
            If the vector has no axis 1.
        TypeError
            If `dim` is not an integer.
        ValueError
            If the shape of axis 1 of the vector is lower than dim.

        Examples
        --------
        >>> v = Vector([[1,2],[0,2],[3,0]])
        >>> v.component(0)
        Vector([1, 0, 3])
        >>> v.component(1)
        Vector([2, 2, 0])
        """

        if len(self.shape) < 2:
            raise TypeError("Operation not supperted on simple vectors")
        if not isinstance(dim, int):
            raise TypeError("Parameter 'dim' must be an integer")
        if self.shape[1] < dim + 1:
            raise ValueError(f"Vector has not component {dim}")
        return self[:, dim].view(Vector)

    @staticmethod
    def create(*args, **kwargs) -> Vector:
        """
        .. deprecated:: 0.10.0
            :func:`Vector.create` will be removed in a future version, use
            :class:`Vector` constructor instead.

        Creates a new vector.

        Returns
        -------
        Vector
            Vector created
        """

        warnings.warn(
            "Vector.create is deprecated and it will be removed in a future version, "
            "use Vector constructor instead.",
            DeprecationWarning,
        )
        return np.array(*args, **kwargs).view(Vector)



"""
CSV traj serializer
"""

from __future__ import annotations

import csv

import numpy as np

import yupi._differentiation as diff
from yupi import Trajectory
from yupi.core.serializers.serializer import Serializer


class CSVSerializer(Serializer):
    """
    Handles trajectory files in JSON format.
    """

    @staticmethod
    def save(
        traj: Trajectory, file_name: str, overwrite: bool = False, **kwargs
    ) -> None:
        """
        Writes a trajectory to a file.

        Parameters
        ----------
        traj : Trajectory
            The trajectory to write to the file.
        file_name : str
            The name of the file to write.
        overwrite : bool
            If True, overwrites the file if it already exists.
        kwargs
            Additional arguments to pass to the ``open`` function.

            Encoding is set to UTF-8 as default.
        """
        CSVSerializer.check_save_path(file_name, overwrite=overwrite, extension=".csv")

        kwargs["encoding"] = kwargs.get("encoding", "utf-8")
        with open(file_name, "w", newline="", **kwargs) as traj_file:
            writer = csv.writer(traj_file, delimiter=",")
            dt = traj.dt if traj.dt_std == 0 else None

            diff_method = Trajectory.general_diff_est.get(
                "method", diff.DiffMethod.LINEAR_DIFF
            )
            diff_win = Trajectory.general_diff_est.get(
                "window_type", diff.WindowType.CENTRAL
            )
            accuracy = Trajectory.general_diff_est.get("accuracy", 1)
            method = traj.diff_est.get("method", diff_method).value
            window = traj.diff_est.get("window_type", diff_win).value
            accuracy = traj.diff_est.get("accuracy", accuracy)

            writer.writerow([traj.traj_id, dt, traj.dim])
            writer.writerow([method, window, accuracy])
            writer.writerows(np.hstack([p, t]) for p, t in zip(traj.r, traj.t))

    @staticmethod
    def load(file_name: str, **kwargs) -> Trajectory:
        """
        Loads a trajectory from a file.

        Parameters
        ----------
        file_name : str
            The name of the file to loaded.
        kwargs : dict
            Additional keyword arguments.

            Encoding is set to UTF-8 as default.

        Returns
        -------
        Trajectory
            The trajectory loaded from the file.
        """

        CSVSerializer.check_load_path(file_name, extension=".csv")

        kwargs["encoding"] = kwargs.get("encoding", "utf-8")
        with open(file_name, "r", **kwargs) as traj_file:
            reader = csv.reader(traj_file, delimiter=",")

            traj_id, dt, dim = next(reader)
            dt = None if not dt else float(dt)
            dim = None if not dim else int(dim)

            method, window, accuracy = list(map(int, next(reader)))
            diff_est = Trajectory.general_diff_est
            diff_est["method"] = diff.DiffMethod(method)
            diff_est["window_type"] = diff.WindowType(window)
            diff_est["accuracy"] = accuracy

            data = np.array([[float(x) for x in row] for row in reader])
            axes = data[:, :dim].T
            t = data[:, dim]
            return Trajectory(axes=axes, t=t, dt=dt, traj_id=traj_id, diff_est=diff_est)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yupi',
 'yupi.core',
 'yupi.core.serializers',
 'yupi.generators',
 'yupi.graphics',
 'yupi.stats',
 'yupi.tracking',
 'yupi.transformations']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.2.0', 'nudged>=0.3.1', 'numpy>=1.16.5', 'opencv-python>=4.4.0']

setup_kwargs = {
    'name': 'yupi',
    'version': '0.11.2',
    'description': 'A package for tracking and analysing objects trajectories',
    'long_description': '<p align="center"><img src="logo.png" alt="Logo"></p>\n\nStanding for *Yet Underused Path Instruments*, **yupi** is a set of tools designed\nfor collecting, generating and processing trajectory data of any kind.\n\n## **What does it offers?**\n\n- **Convert raw data to trajectories** ... *different input manners*\n- **I/O operations with trajectories** ... *json and csv serializers*\n- **Trajectory extraction from video inputs** ... *even with moving camera*\n- **Artificial trajectory generation** ... *several models implemented*\n- **Trajectory basic operations** ... *rotation, shift, scaling, ...*\n- **Trajectory transformations** ... *filters, resamples, ...*\n- **Statistical calculations from trajectories ensembles** ... *turning angles histogram, velocity autocorrelation function, power spectral density, and much more ...*\n- **Results visualization** ... *each statistical observable has a related plot function*\n- **Spacial projection of trajectories** ... *for 2D and 3D trajectories*\n\n## Installation\n\nCurrent recommended installation method is via PyPI:\n\n```cmd\npip install yupi\n```\n\n## Compatibility\n\n- Python 3.7 or later\n- Ubuntu 18.04 or later\n- Windows 7 or later\n- macOS 10.12.6 (Sierra) or later.\n\n## Getting Started\n\nIn the [official documentation](https://yupi.readthedocs.io/en/latest/) there\nare some resources to start using the library: Tutorials, Examples  and a\ndetailed description of the API.\n\n## Examples\n\nCode examples (with additional multimedia resources) can be found in\n[this repository](https://github.com/yupidevs/yupi_examples). Additionally, in\nthe [Examples section](https://yupi.readthedocs.io/en/latest/examples/examples.html)\nof the documentation, you can find the same examples with additional comments\nand expected execution results in order to inspect the examples without actually\nexecuting them.\n',
    'author': 'Gustavo Viera-López',
    'author_email': 'gvieralopez@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/yupidevs/yupi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
This module contains a set of functions for statistical information
extraction from a trajectory collection.

All the resources of this module should be imported directly
from ``yupi.stats``.
"""

from yupi.stats._stats import (
    collect,
    collect_at_step,
    collect_at_time,
    collect_step_lagged,
    collect_time_lagged,
    kurtosis,
    kurtosis_reference,
    msd,
    msd_ensemble,
    msd_time,
    psd,
    speed_ensemble,
    turning_angles_ensemble,
    vacf,
    vacf_ensemble,
    vacf_time,
)

__all__ = [
    "turning_angles_ensemble",
    "collect_at_step",
    "collect_at_time",
    "collect_step_lagged",
    "collect_time_lagged",
    "collect",
    "speed_ensemble",
    "msd_ensemble",
    "msd_time",
    "vacf_ensemble",
    "msd",
    "vacf_time",
    "vacf",
    "kurtosis",
    "kurtosis_reference",
    "psd",
]



"""
This module constains a set of functions dadicated to plot
trajectories and extracted statistical data from ``yupi.stats``.

All the resources of this module should be imported directly
from ``yupi.graphics``.
"""

from yupi.graphics._stats_plots import (
    plot_angles_hist,
    plot_hist,
    plot_hists,
    plot_kurtosis,
    plot_msd,
    plot_psd,
    plot_speed_hist,
    plot_vacf,
)
from yupi.graphics._style import DOTTED, LINE, LINE_DOTTED
from yupi.graphics._trajs_plots import plot_2D, plot_2d, plot_3D, plot_3d, plot_vs_time

__all__ = [
    "plot_angles_hist",
    "plot_kurtosis",
    "plot_msd",
    "plot_vacf",
    "plot_speed_hist",
    "plot_2D",
    "plot_3D",
    "plot_2d",
    "plot_3d",
    "plot_vs_time",
    "plot_psd",
    "plot_hist",
    "plot_hists",
    "LINE",
    "LINE_DOTTED",
    "DOTTED",
]
